# -*- coding: utf-8 -*-
# file: auto_bench.py
# time: 11:54 14/04/2024
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2024. All Rights Reserved.

import os
import time
import warnings

import findfile
import torch
from metric_visualizer import MetricVisualizer

from transformers import TrainingArguments, Trainer as HFTrainer
from ...src.abc.abstract_tokenizer import OmniTokenizer
from ...src.lora.lora_model import OmniLoraModel
from ...src.misc.utils import (
    seed_everything,
    fprint,
    load_module_from_path,
    check_bench_version,
    clean_temp_checkpoint,
)
from ...src.trainer.trainer import Trainer
from ...src.trainer.accelerate_trainer import AccelerateTrainer
from ...src.utility.dataset_hub.dataset_hub import download_benchmark
from ...auto.config.auto_config import AutoConfig
from ... import __version__ as omnigenbench_version


class AutoBench:
    """
    This class provides a comprehensive framework for evaluating genomic models
    across multiple benchmarks and tasks. It handles loading benchmarks, models,
    tokenizers, and running evaluations with proper metric tracking and result
    visualization.

    AutoBench supports various evaluation scenarios including:

    - Single model evaluation across multiple benchmarks
    - Multi-seed evaluation for robustness testing
    - Different trainer backends (native, accelerate, huggingface)
    - Automatic metric visualization and result tracking

    Attributes:
        benchmark (str): The name or path of the benchmark to use.
        model_name_or_path (str): The name or path of the model to evaluate.
        tokenizer: The tokenizer to use for evaluation.
        autocast (str): The autocast precision to use ('fp16', 'bf16', etc.).
        overwrite (bool): Whether to overwrite existing evaluation results.
        trainer (str): The trainer to use ('native', 'accelerate', 'hf_trainer').
        mv_path (str): Path to the metric visualizer file.
        mv (MetricVisualizer): The metric visualizer instance.
        bench_metadata: Metadata about the benchmark configuration.
    """

    def __init__(
        self,
        benchmark,
        model_name_or_path,
        tokenizer=None,
        **kwargs,
    ):
        """
        Initializes the AutoBench instance.

        Args:
            benchmark (str): The name or path of the benchmark to use.
            model_name_or_path (str): The name or path of the model to evaluate.
            tokenizer: The tokenizer to use. If None, it will be loaded from the model path.
            **kwargs: Additional keyword arguments.
                - autocast (str): The autocast precision to use ('fp16', 'bf16', etc.).
                  Defaults to 'fp16'.
                - overwrite (bool): Whether to overwrite existing evaluation results.
                  Defaults to False.
                - trainer (str): The trainer to use ('native', 'accelerate', 'hf_trainer').
                  Defaults to 'native'.

        Example:
            >>> # Initialize with a benchmark and model
            >>> bench = AutoBench("RGB", "model_name")

            >>> # Initialize with custom settings
            >>> bench = AutoBench("RGB", "model_name",
            ...                   autocast="bf16", trainer="accelerate")
        """
        self.benchmark = benchmark.rstrip("/")
        self.autocast = kwargs.pop("autocast", "fp16")
        self.overwrite = kwargs.pop("overwrite", False)
        self.trainer = kwargs.pop("trainer", "native")

        self.model_name_or_path = model_name_or_path
        self.tokenizer = tokenizer
        if isinstance(self.model_name_or_path, str):
            self.model_name_or_path = self.model_name_or_path.rstrip("/")
            self.model_name = self.model_name_or_path.split("/")[-1]
        else:
            self.model_name = self.model_name_or_path.__class__.__name__
        if isinstance(tokenizer, str):
            self.tokenizer = tokenizer.rstrip("/")
        os.makedirs("./autobench_evaluations", exist_ok=True)
        time_str = time.strftime("%Y%m%d_%H%M%S", time.localtime())
        mv_name = f"{benchmark}-{self.model_name}"
        self.mv_path = f"./autobench_evaluations/{mv_name}-{time_str}.mv"

        mv_paths = findfile.find_files(
            "./autobench_evaluations",
            and_key=[benchmark, self.model_name, ".mv"],
        )
        if mv_paths and not self.overwrite:
            self.mv = MetricVisualizer.load(mv_paths[-1])
            self.mv.summary(round=4)
        else:
            self.mv = MetricVisualizer(self.mv_path)

        self.benchmark = download_benchmark(self.benchmark)

        # Import benchmark list
        self.bench_metadata = load_module_from_path(
            f"bench_metadata", f"{self.benchmark}/metadata.py"
        )
        if hasattr(self.bench_metadata, "__omnigenbench_version__"):
            fprint(
                "Benchmark metadata version:",
                self.bench_metadata.__omnigenbench_version__,
            )
            check_bench_version(
                self.bench_metadata.__omnigenbench_version__, omnigenbench_version
            )
        elif hasattr(self.bench_metadata, "__omnigenome_version__"):
            fprint(
                "Benchmark metadata version:",
                self.bench_metadata.__omnigenome_version__,
            )
            check_bench_version(
                self.bench_metadata.__omnigenome_version__, omnigenbench_version
            )

        fprint("Loaded benchmarks: ", self.bench_metadata.bench_list)
        self.bench_info()

    def bench_info(self):
        """
        Prints and returns information about the current benchmark setup.

        Returns:
            str: A string containing benchmark information.

        Example:
            >>> info = bench.bench_info()
            >>> print(info)
        """
        info = f"Benchmark Root: {self.benchmark}\n"
        info += f"Benchmark List: {self.bench_metadata.bench_list}\n"
        info += f"Model Name or Path: {self.model_name}\n"
        info += f"Tokenizer: {self.tokenizer}\n"
        info += f"Metric Visualizer Path: {self.mv_path}\n"
        info += f"BenchConfig Details: {self.bench_metadata}\n"
        fprint(info)
        return info

    def run(self, **kwargs):
        """
        Runs the benchmarking process. This method iterates through the tasks in the benchmark, loads the corresponding
        configurations, initializes the model, tokenizer, and datasets, and then
        trains and evaluates the model.

        Args:
            **kwargs: Additional keyword arguments that will override the default
                     parameters in the benchmark configuration.

        Example:
            >>> # Run benchmarking with default settings
            >>> bench.run()
            >>> # Run with custom parameters
            >>> bench.run(learning_rate=1e-4, batch_size=16)
        """
        bs_scale = kwargs.pop("bs_scale", 1)
        # Import benchmark config
        for _, bench in enumerate(self.bench_metadata.bench_list):
            _kwargs = kwargs.copy()
            clean_temp_checkpoint(1)  # clean temp checkpoint older than 1 day
            fprint(
                ">" * 80,
                f"\nRunning evaluation for task: {bench}",
                "Progress: ",
                _ + 1,
                "/",
                len(self.bench_metadata.bench_list),
                f"{(_ + 1) * 100 / len(self.bench_metadata.bench_list)}%",
            )
            bench_config_path = findfile.find_file(
                self.benchmark,
                and_key=f"{self.benchmark}.{bench}.config".split("."),
            )
            config = load_module_from_path("config", bench_config_path)
            bench_config = None
            for attr_name in dir(config):
                attr = getattr(config, attr_name)
                if isinstance(
                    attr, AutoConfig
                ):  # Check if it is an instance of AutoConfig
                    bench_config = attr
            if bench_config is None:
                raise ValueError(
                    f"Could not find AutoConfig instance in {bench_config_path}"
                )
            fprint(f"Loaded config for {bench} from {bench_config_path}")
            fprint(bench_config)

            # Init Tokenizer and Model
            if not self.tokenizer:
                tokenizer = OmniTokenizer.from_pretrained(
                    self.model_name_or_path,
                    trust_remote_code=bench_config.get("trust_remote_code", True),
                    **bench_config,
                )
            else:
                tokenizer = self.tokenizer

            for key, value in _kwargs.items():
                if key in bench_config:
                    fprint(
                        "Override",
                        key,
                        "with",
                        value,
                        "according to the input kwargs",
                    )
                    bench_config.update({key: value})

                else:
                    warnings.warn(
                        f"kwarg: {key} not found in bench_config while setting {key} = {value}"
                    )
                    bench_config.update({key: value})

            for key, value in bench_config.items():
                if key in bench_config and key in _kwargs:
                    _kwargs.pop(key)

            if not isinstance(bench_config["seeds"], list):
                bench_config["seeds"] = [bench_config["seeds"]]

            random_seeds = bench_config["seeds"]

            for seed in random_seeds:
                batch_size = (
                    bench_config["batch_size"] if "batch_size" in bench_config else 8
                ) * bs_scale

                record_name = f"{self.benchmark}-{bench}-{self.model_name}".split("/")[
                    -1
                ]

                # check if the record exists
                if record_name in self.mv.transpose() and len(
                    list(self.mv.transpose()[record_name].values())[0]
                ) >= len(random_seeds):
                    continue

                seed_everything(seed)
                if self.model_name_or_path:
                    model_cls = bench_config["model_cls"]
                    model = model_cls(
                        self.model_name_or_path,
                        tokenizer=tokenizer,
                        label2id=bench_config.label2id,
                        num_labels=bench_config["num_labels"],
                        trust_remote_code=True,
                        ignore_mismatched_sizes=True,
                    )
                else:
                    raise ValueError(
                        "model_name_or_path is not specified. Please provide a valid model name or path."
                    )

                fprint(f"\n{model}")

                if kwargs.get("lora_config", {}) or kwargs.get("lora", True):
                    fprint(
                        "Applying LoRA to the model with config:",
                        kwargs.get("lora_config", {}) or "Default Config",
                    )
                    model = OmniLoraModel(model, **kwargs.get("lora_config", {}))

                # Init Trainer
                dataset_cls = bench_config["dataset_cls"]

                if hasattr(model.config, "max_position_embeddings"):
                    max_length = min(
                        bench_config["max_length"],
                        model.config.max_position_embeddings,
                    )
                else:
                    max_length = bench_config["max_length"]

                train_set = dataset_cls(
                    dataset_name_or_path=bench_config["train_file"],
                    tokenizer=tokenizer,
                    label2id=bench_config["label2id"],
                    max_length=max_length,
                    structure_in=bench_config.get("structure_in", False),
                    max_examples=bench_config.get("max_examples", None),
                    shuffle=bench_config.get("shuffle", True),
                    drop_long_seq=bench_config.get("drop_long_seq", False),
                    **_kwargs,
                )
                test_set = dataset_cls(
                    dataset_name_or_path=bench_config["test_file"],
                    tokenizer=tokenizer,
                    label2id=bench_config["label2id"],
                    max_length=max_length,
                    structure_in=bench_config.get("structure_in", False),
                    max_examples=bench_config.get("max_examples", None),
                    shuffle=False,
                    drop_long_seq=bench_config.get("drop_long_seq", False),
                    **_kwargs,
                )
                valid_set = dataset_cls(
                    dataset_name_or_path=bench_config["valid_file"],
                    tokenizer=tokenizer,
                    label2id=bench_config["label2id"],
                    max_length=max_length,
                    structure_in=bench_config.get("structure_in", False),
                    max_examples=bench_config.get("max_examples", None),
                    shuffle=False,
                    drop_long_seq=bench_config.get("drop_long_seq", False),
                    **_kwargs,
                )

                if self.trainer == "hf_trainer":
                    # Set up HuggingFace Trainer
                    hf_kwargs = {
                        k: v
                        for k, v in kwargs.items()
                        if hasattr(TrainingArguments, k) and k != "output_dir"
                    }
                    training_args = TrainingArguments(
                        output_dir=f"autobench_evaluations/{self.model_name}-{bench}",
                        num_train_epochs=hf_kwargs.pop(
                            "num_train_epochs", bench_config["epochs"]
                        ),
                        per_device_train_batch_size=hf_kwargs.pop(
                            "batch_size", batch_size
                        ),
                        per_device_eval_batch_size=hf_kwargs.pop(
                            "batch_size", batch_size
                        ),
                        gradient_accumulation_steps=hf_kwargs.pop(
                            "gradient_accumulation_steps", 1
                        ),
                        learning_rate=hf_kwargs.pop("learning_rate", 2e-5),
                        weight_decay=hf_kwargs.pop("weight_decay", 0),
                        eval_strategy=hf_kwargs.pop("eval_strategy", "epoch"),
                        save_strategy=hf_kwargs.pop("save_strategy", "epoch"),
                        fp16=hf_kwargs.pop("fp16", True),
                        remove_unused_columns=False,
                        label_names=["labels"],
                        **hf_kwargs,
                    )

                    valid_set = valid_set if len(valid_set) else test_set

                    if len(bench_config["compute_metrics"]) > 1:
                        fprint(
                            "Multiple metrics not supported by HFTrainer, using the first one metric only."
                        )
                    trainer = HFTrainer(
                        model=model,
                        args=training_args,
                        train_dataset=train_set,
                        eval_dataset=valid_set,
                        compute_metrics=(
                            bench_config["compute_metrics"][0]
                            if isinstance(bench_config["compute_metrics"], list)
                            else bench_config["compute_metrics"]
                        ),
                    )

                    # Train and evaluate
                    eval_result = trainer.evaluate(
                        valid_set if len(valid_set) else test_set
                    )
                    print(eval_result)
                    train_result = trainer.train()
                    eval_result = trainer.evaluate()
                    test_result = trainer.evaluate(
                        test_set if len(test_set) else valid_set
                    )

                    metrics = {
                        "train": train_result.metrics,
                        "eval": eval_result,
                        "test": test_result,
                    }
                    fprint(metrics)
                else:
                    optimizer = torch.optim.AdamW(
                        filter(lambda p: p.requires_grad, model.parameters()),
                        lr=(
                            bench_config["learning_rate"]
                            if "learning_rate" in bench_config
                            else 2e-5
                        ),
                        weight_decay=(
                            bench_config["weight_decay"]
                            if "weight_decay" in bench_config
                            else 0
                        ),
                    )
                    if self.trainer == "accelerate":
                        trainer_cls = AccelerateTrainer
                    else:
                        trainer_cls = Trainer
                    fprint(f"Using Trainer: {trainer_cls}")
                    trainer = trainer_cls(
                        model=model,
                        train_dataset=train_set,
                        eval_dataset=valid_set,
                        test_dataset=test_set,
                        batch_size=batch_size,
                        patience=(
                            bench_config["patience"]
                            if "patience" in bench_config
                            else 3
                        ),
                        epochs=bench_config["epochs"],
                        gradient_accumulation_steps=bench_config.get(
                            "gradient_accumulation_steps", 1
                        ),
                        optimizer=optimizer,
                        loss_fn=(
                            bench_config["loss_fn"]
                            if "loss_fn" in bench_config
                            else None
                        ),
                        compute_metrics=bench_config["compute_metrics"],
                        seed=seed,
                        autocast=self.autocast,
                        **_kwargs,
                    )
                    metrics = trainer.train()

                    predictions = trainer.predictions

                    if bench_config.get("save_predictions", False):
                        os.makedirs(f"predictions/{bench}", exist_ok=True)
                        import numpy as np

                        for split in predictions.keys():
                            with open(
                                f"predictions/{bench}/{split}.npy",
                                "wb",
                            ) as f:
                                np.save(f, predictions[split])

                    if metrics:
                        for key, value in metrics["test"][-1].items():
                            try:
                                value = float(value)
                            except:
                                pass  # ignore non-float values
                            self.mv.log(f"{record_name}", f"{key}", value)
                        # for key, value in metrics['test'][-1].items():
                        #     self.mv.log(f'{record_name}', f'test_{key}', value)
                        # for i, valid_metrics in enumerate(metrics["valid"]):
                        #     for key, value in valid_metrics.items():
                        #         self.mv.log(f'{record_name}', f'valid_epoch_{i}_{key}', value)

                        self.mv.summary(round=4)
                        self.mv.dump(self.mv_path)
                        self.mv.to_csv(self.mv_path.replace(".mv", ".csv"))
                    del model, trainer, optimizer
                    torch.cuda.empty_cache()



# -*- coding: utf-8 -*-
# file: ism_explainer.py
# time: 2025-08-27 15:19
# author: Shasha Zhou <sz484@exeter.ac.uk>
# Copyright (C) 2020-2025. All Rights Reserved.

from ...abc.abstract_explainer import AbstractExplainer
from ...misc.utils import fprint
import numpy as np
from ...model.classification.model import (
    OmniModelForSequenceClassification,
    OmniModelForTokenClassification,
)
from ...model.regression.model import (
    OmniModelForSequenceRegression,
    OmniModelForTokenRegression,
)
from typing import List
import torch


class ISMExplainer(AbstractExplainer):
    """Explains model predictions using the ISM method.

    ISM (In Silico MAVE) is a method that uses in-silico mutagenesis to generate a dataset, which is then
    used to train a simpler, interpretable surrogate model. From this surrogate model, additive (first-order) or pairwise (second-order) feature attributions
    can be extracted.
    """

    def __init__(self, model):
        """Initialize the ISMExplainer.

        Args:
            model: The trained model (classification or regression) to explain.
        """
        super().__init__(model)
        self.token_to_id = {}
        self.alphabet = []
        self.num_tokens = 0

    def explain(self, sequence: str, task_idx: int = 0, batch_size: int = 32, **kwargs):
        """Generate ISM explanations for a given sequence.

        Args:
            sequence (str): The input sequence to explain.
            task_idx (int, optional): Target task index (for multi-task models). Defaults to 0.
            batch_size (int, optional): Batch size for prediction. Defaults to 32.
            **kwargs: Additional arguments passed to the model's `predict` method.

        Returns:
            np.ndarray: A score matrix of shape (len(sequence), num_tokens).
                - Each row corresponds to a position in the input sequence.
                - Each column corresponds to a token in the alphabet.
                - Entry [i, j] is the change in model output when the i-th token
                  is mutated to alphabet[j]. For the original base, its value is
                  set as the negative sum of the other mutations at that position.

        Raises:
            ValueError: If the task index is invalid or model type unsupported.
        """
        fprint("Starting ISM explanation")
        # initialize the alphabet and id_to_token
        unique_chars = []
        for ch in sequence:
            if ch not in unique_chars:
                unique_chars.append(ch)
        self.token_to_id = {ch: idx for idx, ch in enumerate(unique_chars)}
        self.alphabet = unique_chars
        self.num_tokens = len(unique_chars)

        # in silico MAVE data generation
        fprint("Generating in silico MAVE data...")
        mut_seqs, mut_positions, mut_bases = self._generate_in_silico_mave(sequence)

        y_ref = self.model.predict(sequence, **kwargs)["logits"][0]
        if task_idx >= len(y_ref):
            raise ValueError(f"Task index {task_idx} is out of range for the model.")
        y_ref = y_ref[task_idx]

        # calculate the output of the model for the mutated sequences
        y_mut = self._mut_predictor(
            mut_seqs, task_idx=task_idx, batch_size=batch_size, **kwargs
        )

        scores = np.zeros(len(mut_seqs, len(self.alphabet)))

        for k, (pos, base) in enumerate(zip(mut_positions, mut_bases)):
            base_idx = self.alphabet.index(base)
            scores[k, base_idx] = y_mut[k] - y_ref

        for i in range(len(sequence)):
            orig_base = sequence[i]
            orig_idx = self.alphabet.index(orig_base)
            other_sum = np.sum(
                scores[i, j] for j in range(len(self.alphabet)) if j != orig_idx
            )
            scores[i, orig_idx] = -other_sum

        return scores

    def _generate_in_silico_mave(self, sequence: str):
        """(Private) Generate single-mutation variants of the sequence.

        This function creates all possible point mutations of the input sequence,
        by replacing each position with every other token in the alphabet.

        Args:
            sequence (str): Original input sequence.

        Returns:
            Tuple[List[str], List[int], List[str]]:
                - mut_seqs: List of mutated sequences (strings).
                - mut_positions: List of mutated positions (ints).
                - mut_bases: List of substituted bases/tokens (str).
        """
        L = len(list(sequence))

        mut_seqs = []
        mut_positions = []
        mut_bases = []

        for i in range(L):
            for j, b in enumerate(self.alphabet):
                if b == sequence[i]:
                    continue
                mut_seq = list(sequence)
                mut_seq[i] = b
                mut_seqs.append("".join(mut_seq))
                mut_positions.append(i)
                mut_bases.append(b)

        return mut_seqs, mut_positions, mut_bases

    def _mut_predictor(
        self, mut_seqs: List[str], task_idx: int = 0, batch_size: int = 32, **kwargs
    ):
        """(Private) Predict the model's output for a list of mutated sequences.

        This function uses the model to predict the output for each mutated sequence.

        Args:
            mut_seqs (List[str]): List of mutated sequences.
            task_idx (int, optional): Target task index (for multi-task models). Defaults to 0.
            batch_size (int, optional): Batch size for prediction. Defaults to 32.
            **kwargs: Additional arguments passed to the model's `predict` method.

        Returns:
            np.ndarray: Model predictions for each mutated sequence.

        Raises:
            ValueError: If the task index is invalid or model type unsupported.
        """
        if isinstance(self.model, OmniModelForSequenceClassification) or isinstance(
            self.model, OmniModelForTokenClassification
        ):
            y_mut = []
            for i in range(0, len(mut_seqs), batch_size):
                batch = mut_seqs[i : i + batch_size]
                outputs = self.model.predict(batch, **kwargs)["logits"]
                if task_idx >= len(outputs[0]):
                    raise ValueError(
                        f"Task index {task_idx} is out of range for the model."
                    )
                y_mut.append(outputs[:, task_idx])
            y_mut = torch.cat(y_mut, dim=0).detach().cpu().numpy()
        elif isinstance(self.model, OmniModelForSequenceRegression) or isinstance(
            self.model, OmniModelForTokenRegression
        ):
            y_mut = []
            for i in range(0, len(mut_seqs), batch_size):
                batch = mut_seqs[i : i + batch_size]
                outputs = self.model.predict(batch, **kwargs)["predictions"]
                y_mut.append(outputs[:, task_idx])
            y_mut = torch.cat(y_mut, dim=0).detach().cpu().numpy()
        else:
            raise ValueError(
                f"Model type {type(self.model)} not supported for ISM explainer."
            )
        return y_mut



# -*- coding: utf-8 -*-
# file: abstract_dataset.py
# time: 14:13 06/04/2024
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2024. All Rights Reserved.
import os
import random
import warnings
import zipfile
import findfile
import requests
from collections import Counter

import numpy as np
import torch
import tqdm

from transformers import BatchEncoding

from ..misc.utils import fprint, env_meta_info, RNA2StructureCache


def covert_input_to_tensor(data):
    """
    This function traverses through nested dictionaries and lists, converting
    numerical values to PyTorch tensors while preserving the structure.

    Args:
        data (list or dict): A list or dictionary containing data samples.

    Returns:
        list or dict: The data structure with numerical values converted to tensors.

    Example:
        >>> data = [{'input_ids': [1, 2, 3], 'labels': [0]}]
        >>> tensor_data = covert_input_to_tensor(data)
        >>> print(type(tensor_data[0]['input_ids']))  # <class 'torch.Tensor'>
    """
    for d in data:
        if isinstance(d, dict) or isinstance(d, BatchEncoding):
            for key, value in d.items():
                try:
                    if not isinstance(value, torch.Tensor):
                        d[key] = torch.tensor(value)
                except Exception as e:
                    pass
        elif isinstance(d, list):
            for value in d:
                covert_input_to_tensor(value)
            covert_input_to_tensor(d)

    return data


class OmniDict(dict):
    """
    This class extends the standard Python dictionary to provide a convenient
    method for moving all tensor values to a specific device (CPU/GPU).
    """

    def __init__(self, *args, **kwargs):
        super(OmniDict, self).__init__(*args, **kwargs)

    def to(self, device):
        """
        Moves all tensor values in the dictionary to the specified device.

        Args:
            device (str or torch.device): The target device (e.g., 'cuda:0' or 'cpu').

        Returns:
            OmniDict: The dictionary itself, with tensors moved to the new device.

        Example:
            >>> data = OmniDict({'input_ids': torch.tensor([1, 2, 3])})
            >>> data.to('cuda:0')  # Moves tensors to GPU
        """
        for key, value in self.items():
            if isinstance(value, torch.Tensor):
                self[key] = value.to(device)
        return self


class OmniDataset(torch.utils.data.Dataset):
    """
    A unified interface for genomic datasets in the OmniGenome
    framework. It handles data loading, preprocessing, tokenization, and provides
    a PyTorch-compatible dataset interface.

    The class supports various data formats and can handle different types of
    genomic tasks including classification, regression, and token-level tasks.

    Attributes:
        tokenizer: The tokenizer to use for processing sequences.
        max_length (int): The maximum sequence length for tokenization.
        label2id (dict): Mapping from labels to integer IDs.
        id2label (dict): Mapping from integer IDs to labels.
        shuffle (bool): Whether to shuffle the data.
        structure_in (bool): Whether to include secondary structure information.
        drop_long_seq (bool): Whether to drop sequences longer than max_length.
        metadata (dict): Metadata about the dataset including version info.
        rna2structure (RNA2StructureCache): Cache for RNA structure predictions.
    """

    def __init__(
        self, dataset_name_or_path=None, tokenizer=None, max_length=None, **kwargs
    ):
        """
        Initializes the dataset.

        Args:
            dataset_name_or_path (str or list): Path to the data file or a list of paths.
            tokenizer: The tokenizer to use for processing sequences.
            max_length (int, optional): The maximum sequence length.
            **kwargs: Additional keyword arguments.
                - label2id (dict): A mapping from labels to integer IDs.
                - shuffle (bool): Whether to shuffle the data. Defaults to True.
                - structure_in (bool): Whether to include secondary structure
                  information. Defaults to False.
                - drop_long_seq (bool): Whether to drop sequences longer than
                  max_length. Defaults to False.
                - dataset_url (str): URL to download dataset if not found locally.
                - cache_dir (str): Directory to cache downloaded datasets.

        Example:
            >>> # Initialize with a single data file
            >>> dataset = OmniDataset("data.json", tokenizer, max_length=512)

            >>> # Initialize with label mapping
            >>> dataset = OmniDataset("data.json", tokenizer,
            ...                       label2id={"A": 0, "B": 1})

            >>> # Initialize with automatic dataset download
            >>> dataset = OmniDataset("data.csv", tokenizer,
            ...                       dataset_url="https://example.com/data.zip")
        """
        super(OmniDataset, self).__init__()
        if not dataset_name_or_path and kwargs.get("data_source", None):
            dataset_name_or_path = kwargs.get("data_source", None)
        if not dataset_name_or_path:
            raise ValueError("Please provide dataset_name_or_path")

        self.metadata = env_meta_info()
        self.dataset_info = None  # Store dataset_info separately from metadata
        self.tokenizer = tokenizer
        self.label2id = kwargs.get("label2id", None)
        self.shuffle = kwargs.get("shuffle", True)
        self.structure_in = kwargs.get("structure_in", False)
        self.drop_long_seq = kwargs.get("drop_long_seq", False)
        self.force_padding = kwargs.get("force_padding", True)
        if self.structure_in and not hasattr(self, "rna2structure"):
            self.rna2structure = RNA2StructureCache()

        if self.label2id is not None:
            self.id2label = {v: k for k, v in self.label2id.items()}
        else:
            fprint(
                "No label2id provided, something wrong may happen. "
                "label2id indicates the mapping from label to indices. "
                "e.g., {'positive': 1, 'negative': 0} for binary classification."
            )

        if max_length is not None:
            fprint(
                f"Detected max_length={max_length} in the dataset, using it as the max_length."
            )
            self.max_length = max_length
        elif (
            hasattr(self.tokenizer, "max_length")
            and self.tokenizer.max_length is not None
        ):
            fprint(
                f"Detected max_length={self.tokenizer.max_length} from the tokenizer."
            )
            self.max_length = self.tokenizer.max_length
        else:
            fprint(f"No max_length detected, using default max_length=512.")
            self.max_length = 512

        self.tokenizer.max_length = self.max_length
        self.examples = []
        self.data = []

        if dataset_name_or_path is not None:
            # Check if dataset needs to be downloaded
            fprint(f"Loading data from {dataset_name_or_path}...")
            self.load_data_source(dataset_name_or_path, **kwargs)
            # Try to load dataset_info.json from the same directory
            self._load_dataset_info(dataset_name_or_path)
            self._preprocessing()
            if self.tokenizer is not None:
                for example in tqdm.tqdm(self.examples):
                    if hasattr(self.tokenizer, "max_length"):
                        self.tokenizer.max_length = self.max_length
                    else:
                        self.tokenizer.base_tokenizer.max_length = self.max_length

                    import inspect

                    new_args = {}
                    tokenization_args = inspect.getfullargspec(
                        self.tokenizer.encode
                    ).args
                    for key in kwargs:
                        if key in tokenization_args:
                            new_args[key] = kwargs[key]
                    prepared_input = self.prepare_input(example, **new_args)

                    if not prepared_input:
                        continue

                    if (
                        self.drop_long_seq
                        and len(prepared_input["input_ids"]) > self.max_length
                    ):
                        fprint(
                            f"Dropping sequence {example['sequence']} due to length > {self.max_length}"
                        )
                    else:

                        if isinstance(prepared_input, BatchEncoding):
                            prepared_input = [prepared_input]

                        # Squeeze the batch dimension if it exists
                        for item in prepared_input:
                            for key, value in item.items():
                                item[key] = value.squeeze(0)
                            self.data.append(item)

                self._postprocessing()
                self._pad_and_truncate()

    def get_dataloader(
        self, batch_size=16, shuffle=None, num_workers=0, pin_memory=None, **kwargs
    ):
        """
        Creates a PyTorch DataLoader for this dataset.

        Args:
            batch_size (int): Batch size for the DataLoader.
            shuffle (bool): Whether to shuffle the data. If None, uses self.shuffle.
            num_workers (int): Number of worker processes for data loading.
            pin_memory (bool): Whether to pin memory. If None, auto-detects based on CUDA availability.
            **kwargs: Additional arguments passed to DataLoader.

        Returns:
            torch.utils.data.DataLoader: A DataLoader for this dataset.
        """
        if shuffle is None:
            shuffle = self.shuffle

        if pin_memory is None:
            pin_memory = torch.cuda.is_available()

        from torch.utils.data import DataLoader

        return DataLoader(
            self,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=pin_memory,
            **kwargs,
        )

    @classmethod
    def from_hub(
        cls,
        dataset_name_or_path,
        tokenizer,
        splits=None,
        max_length=None,
        cache_dir=None,
        **kwargs,
    ):
        """
        Create OmniDataset instances from HuggingFace Hub or local directory.

        This method supports loading datasets from:
        1. OmniGenBench Hub on HuggingFace (downloads if needed)
        2. Local directory containing dataset files

        Args:
            dataset_name_or_path (str): Name of the dataset on HuggingFace Hub, or path to local directory.
            tokenizer: The tokenizer to use for processing sequences.
            splits (list, optional): List of splits to create. Defaults to ['train', 'valid', 'test'].
            max_length (int, optional): Maximum sequence length.
            cache_dir (str, optional): Directory to cache the dataset or look for local files.
            **kwargs: Additional arguments passed to the dataset constructor.

        Returns:
            dict: Dictionary containing datasets for each split.

        Example:
            >>> from omnigenbench import OmniTokenizer, OmniDatasetForSequenceClassification
            >>> tokenizer = OmniTokenizer.from_pretrained("yangheng/OmniGenome-52M")

            >>> # Load from HuggingFace Hub
            >>> datasets = OmniDatasetForSequenceClassification.from_hub(
            ...     "translation_efficiency_prediction",
            ...     tokenizer=tokenizer
            ... )

            >>> # Load from local directory
            >>> datasets = OmniDatasetForSequenceClassification.from_hub(
            ...     "/path/to/local/dataset",
            ...     tokenizer=tokenizer,
            ...     cache_dir="/path/to/local/dataset"
            ... )

            >>> train_loader = datasets['train'].get_dataloader(batch_size=16)
        """
        if splits is None:
            splits = ["train", "valid", "test"]

        # Determine if dataset_name_or_path is a local path or a HuggingFace dataset name
        is_local = os.path.exists(dataset_name_or_path) or (
            cache_dir and os.path.exists(cache_dir)
        )

        if not is_local:
            # Download from HuggingFace if not a local path
            cls._download_dataset_from_hub(dataset_name_or_path, cache_dir)
        else:
            fprint(
                f"Loading dataset from local path: {dataset_name_or_path or cache_dir}"
            )

        # Create datasets for each split
        datasets = {"train": None, "valid": None, "test": None}
        keys_to_search = {
            "train": ["train."],
            "valid": ["val.", "valid.", "dev."],
            "test": ["test."],
        }
        for split in splits:
            if not cache_dir:
                if is_local and os.path.exists(dataset_name_or_path):
                    # If dataset_name_or_path itself is a local directory
                    cache_dir = dataset_name_or_path
                else:
                    # Use default cache directory
                    cache_dir = os.getcwd()

            data_source = findfile.find_files(
                cache_dir,
                or_key=keys_to_search[split],
                exclude_key=[".ipynb", ".py", "md", "txt"],
            )
            if not data_source:
                fprint(
                    f"Warning: No data files found for split '{split}' in {cache_dir}. Skipping this split."
                )
                continue
            else:
                fprint(f"Load data files for split '{split}': {data_source}")

            datasets[split] = cls(
                dataset_name_or_path=data_source,
                tokenizer=tokenizer,
                max_length=max_length,
                split=split,
                **kwargs,
            )

        return datasets

    @classmethod
    def from_huggingface(
        cls,
        dataset_name_or_path,
        tokenizer,
        splits=None,
        max_length=None,
        cache_dir=None,
        **kwargs,
    ):
        """
        Create OmniDataset instances from a HuggingFace dataset.

        .. deprecated::
            `from_huggingface` is deprecated and will be removed in a future version.
            Use `from_huggingface` instead, which supports both HuggingFace Hub and local data sources.

        Args:
            dataset_name_or_path (str): Name of the HuggingFace dataset or base URL.
            tokenizer: The tokenizer to use for processing sequences.
            splits (list, optional): List of splits to create. Defaults to ['train', 'valid', 'test'].
            max_length (int, optional): Maximum sequence length.
            cache_dir (str, optional): Directory to cache the dataset.
            **kwargs: Additional arguments passed to the dataset constructor.

        Returns:
            dict: Dictionary containing datasets for each split.

        Example:
            >>> from omnigenbench import OmniTokenizer,OmniDatasetForSequenceClassification
            >>> tokenizer = OmniTokenizer.from_pretrained("yangheng/OmniGenome-52M")
            >>> datasets = OmniDatasetForSequenceClassification.from_huggingface(
            ...     "translation_efficiency_prediction",
            ...     tokenizer=tokenizer
            ... )
            >>> train_loader = datasets['train'].get_dataloader(batch_size=16)
        """
        warnings.warn(
            "from_huggingface() is deprecated and will be removed in a future version. "
            "Please use from_huggingface() instead, which supports both HuggingFace Hub and local data sources.",
            DeprecationWarning,
            stacklevel=2,
        )

        return cls.from_hub(
            dataset_name_or_path=dataset_name_or_path,
            tokenizer=tokenizer,
            splits=splits,
            max_length=max_length,
            cache_dir=cache_dir,
            **kwargs,
        )

    def print_label_distribution(self):
        """
        Print the distribution of labels for 0-dimensional (scalar) labels.
        This is useful for classification tasks where each sample has a single label.
        """
        # Check if we have scalar labels
        if self.data and "labels" in self.data[0]:
            first_label = self.data[0]["labels"]
            if isinstance(first_label.item(), float):
                return

            if not isinstance(first_label, torch.Tensor) or first_label.ndim == 0:
                # Convert labels to list of integers
                labels = [int(d["labels"]) for d in self.data]

                # Count frequency of each label
                label_counts = Counter(labels)
                total_samples = len(labels)

                # Sort by label value
                sorted_counts = sorted(label_counts.items())

                fprint("\nLabel Distribution:")
                fprint("-" * 40)
                fprint(f"{'Label':<10}\t\t{'Count':<10}\t\t{'Percentage':<10}")
                fprint("-" * 40)

                for label, count in sorted_counts:
                    percentage = (count / total_samples) * 100
                    label_name = (
                        self.id2label[label]
                        if hasattr(self, "id2label")
                        else str(label)
                    )
                    fprint(f"{label_name:<10}\t\t{count:<10}\t\t{percentage:.2f}%")

                fprint("-" * 40)
                fprint(f"Total samples: {total_samples}")
            else:
                fprint(
                    "Warning: This method is only for scalar (0-dimensional) labels."
                )
        else:
            fprint("No labels found in the dataset.")

    def to(self, device):
        """
        Moves all tensor data in the dataset to the specified device.

        Args:
            device (str or torch.device): The target device.

        Returns:
            OmniDataset: The dataset itself.
        """
        for data_item in self.data:
            for key, value in data_item.items():
                if isinstance(value, torch.Tensor):
                    data_item[key] = value.to(device)
        return self

    def _pad_and_truncate(self, pad_value=0):
        """
        Pads and truncates sequences in the dataset to a uniform length.
        The length is determined dynamically based on the longest sequence in the batch,
        up to the `self.max_length` limit, and adjusted to be a multiple of 8.

        Args:
            pad_value (int, optional): The value to use for padding. Defaults to 0.

        Returns:
            list: The padded and truncated data.
        """
        key_lengths = {key: [] for key in self.data[0].keys()}
        for item in self.data:
            for key, value in item.items():
                if not isinstance(value, torch.Tensor):
                    value = torch.as_tensor(value)
                length = value.size(0) if value.ndim > 0 else 0
                key_lengths[key].append(length)

        skip_padding_for_key = {
            key: len(set(lengths)) == 1 for key, lengths in key_lengths.items()
        }

        skipped_keys = [key for key, skip in skip_padding_for_key.items() if skip]
        if len(skipped_keys) == len(self.data[0].keys()) and not self.force_padding:
            fprint(
                "All keys have consistent sequence lengths, skipping padding and truncation."
            )
            return self.data

        if hasattr(self.tokenizer, "pad_token_id"):
            pad_token_id = self.tokenizer.pad_token_id
        else:
            pad_token_id = self.tokenizer.base_tokenizer.pad_token_id

        max_input_length = max(
            [
                torch.sum(data_item["input_ids"] != pad_token_id).item()
                for data_item in self.data
            ]
        )
        max_label_length = max(
            [
                (data_item["labels"].shape[0] if data_item["labels"].ndim >= 1 else 0)
                for data_item in self.data
            ]
        )

        original_max_length = max(max_input_length, max_label_length)
        original_max_length = min(original_max_length, self.max_length)

        remainder = original_max_length % 8
        if remainder != 0:
            adjusted_max_length = original_max_length + (8 - remainder)
            adjusted_max_length = min(adjusted_max_length, self.max_length)
        else:
            adjusted_max_length = original_max_length
        max_length = adjusted_max_length

        first_labels = self.data[0]["labels"]

        label_shape = first_labels.shape
        if len(label_shape) >= 1:
            max_length = max(max_length, self.data[0]["labels"].shape[0])
            label_padding_length = min(max_label_length, max_length)
            max_length = max(max_length, label_padding_length)
        else:
            label_padding_length = 0

        fprint(
            f"Max sequence length updated -> Reset max_length={max_length},"
            f" label_padding_length={label_padding_length}"
        )

        for data_item in self.data:
            for key, value in data_item.items():

                if not isinstance(value, torch.Tensor):
                    value = torch.as_tensor(value)
                dtype = value.dtype
                if "label" in key and (
                    value.dtype == torch.int16 or value.dtype == torch.int32
                ):
                    data_item[key] = value.long()

                if "label" in key:
                    if value.ndim == 0:
                        padding_length = 0
                    else:
                        padding_length = label_padding_length - value.size(0)
                else:
                    padding_length = max_length - value.size(0)

                if padding_length > 0:

                    if key == "input_ids":
                        _pad_value = pad_token_id
                    elif key == "attention_mask":
                        _pad_value = 0
                    elif "ids" in key:
                        _pad_value = 0
                    elif "label" in key:
                        _pad_value = -100
                    elif "ids" in key:
                        _pad_value = pad_token_id
                    else:
                        _pad_value = pad_value

                    if value.ndim == 2:
                        pad_shape = (padding_length, value.size(1))
                    else:
                        pad_shape = (padding_length,)
                    pad_tensor = torch.full(pad_shape, _pad_value, dtype=dtype)
                    data_item[key] = torch.cat([value, pad_tensor], dim=0)
                elif padding_length < 0:
                    data_item[key] = value[:max_length]

                data_item[key] = data_item[key].to(dtype)

        return self.data

    def _load_dataset_info(self, dataset_path):
        """
        Load dataset_info.json from the dataset directory.

        This method searches for dataset_info.json in the directory containing
        the dataset file(s). The dataset_info provides structured metadata about
        the dataset including description, statistics, features, and more.

        Note: This is separate from the 'metadata' attribute used by OmniGenBench
        for model/tokenizer versioning. dataset_info is specifically for dataset
        documentation and characteristics.

        Args:
            dataset_path (str or list): Path to dataset file(s) or directory
        """
        import json
        import os

        # Determine the directory to search
        if isinstance(dataset_path, list):
            # Use the directory of the first file
            search_dir = os.path.dirname(os.path.abspath(dataset_path[0]))
        elif os.path.isdir(dataset_path):
            search_dir = dataset_path
        else:
            search_dir = os.path.dirname(os.path.abspath(dataset_path))

        # Look for dataset_info.json in the directory
        info_path = os.path.join(search_dir, "dataset_info.json")

        if os.path.exists(info_path):
            try:
                with open(info_path, "r", encoding="utf-8") as f:
                    self.dataset_info = json.load(f)
                fprint(f"✓ Loaded dataset_info from: {info_path}")
            except Exception as e:
                fprint(f"Warning: Failed to load dataset_info.json: {e}")
                self.dataset_info = None
        else:
            fprint(f"Note: No dataset_info.json found in {search_dir}")
            self.dataset_info = None

    def info(self, sections=None):
        """
        Print formatted dataset information.

        This method displays dataset_info in a human-readable format. It's separate
        from model/tokenizer metadata and focuses on dataset characteristics.

        Args:
            sections (list, optional): List of section names to print. If None, prints all.
                Available sections: 'basic', 'statistics', 'features', 'splits',
                'preprocessing', 'metrics', 'citation', 'all'

        Example:
            >>> dataset.info()  # Print all sections
            >>> dataset.info(sections=['basic', 'statistics'])  # Print specific sections
        """
        if self.dataset_info is None:
            fprint(
                "No dataset_info available. Load a dataset with dataset_info.json or use load_dataset_info()."
            )
            return

        if sections is None or "all" in sections:
            sections = [
                "basic",
                "statistics",
                "features",
                "splits",
                "preprocessing",
                "metrics",
                "citation",
            ]

        info = self.dataset_info

        fprint("\n" + "=" * 70)
        fprint("DATASET INFORMATION".center(70))
        fprint("=" * 70)

        # Basic Information
        if "basic" in sections:
            fprint("\n📊 BASIC INFORMATION")
            fprint("-" * 70)
            if "dataset_name" in info:
                fprint(f"  Name: {info['dataset_name']}")
            if "dataset_version" in info:
                fprint(f"  Version: {info['dataset_version']}")
            if "description" in info:
                fprint(f"  Description: {info['description']}")
            if "task_type" in info:
                fprint(f"  Task Type: {info['task_type']}")
            if "domain" in info:
                fprint(f"  Domain: {info['domain']}")
            if "source" in info:
                fprint(f"  Source: {info['source']}")
            if "license" in info:
                fprint(f"  License: {info['license']}")

        # Statistics
        if "statistics" in sections and "statistics" in info:
            fprint("\n📈 STATISTICS")
            fprint("-" * 70)
            stats = info["statistics"]
            for key, value in stats.items():
                formatted_key = key.replace("_", " ").title()
                if isinstance(value, dict):
                    fprint(f"  {formatted_key}:")
                    for sub_key, sub_value in value.items():
                        fprint(f"    - {sub_key}: {sub_value}")
                else:
                    fprint(f"  {formatted_key}: {value}")

        # Features
        if "features" in sections and "features" in info:
            fprint("\n🔧 FEATURES")
            fprint("-" * 70)
            features = info["features"]

            if "input" in features:
                fprint("  Input Features:")
                for feat_name, feat_info in features["input"].items():
                    fprint(f"    • {feat_name}")
                    if isinstance(feat_info, dict):
                        if "description" in feat_info:
                            fprint(f"      Description: {feat_info['description']}")
                        if "type" in feat_info:
                            fprint(f"      Type: {feat_info['type']}")

            if "output" in features:
                fprint("  Output Features:")
                for feat_name, feat_info in features["output"].items():
                    fprint(f"    • {feat_name}")
                    if isinstance(feat_info, dict) and "description" in feat_info:
                        fprint(f"      Description: {feat_info['description']}")

        # Data Splits
        if "splits" in sections and "data_splits" in info:
            fprint("\n📂 DATA SPLITS")
            fprint("-" * 70)
            for split_name, split_info in info["data_splits"].items():
                fprint(f"  {split_name.title()}:")
                if isinstance(split_info, dict):
                    for key, value in split_info.items():
                        fprint(f"    - {key}: {value}")

        # Preprocessing
        if "preprocessing" in sections and "preprocessing" in info:
            fprint("\n⚙️  PREPROCESSING")
            fprint("-" * 70)
            preproc = info["preprocessing"]
            for key, value in preproc.items():
                formatted_key = key.replace("_", " ").title()
                if isinstance(value, list):
                    fprint(f"  {formatted_key}: {', '.join(value)}")
                else:
                    fprint(f"  {formatted_key}: {value}")

        # Evaluation Metrics
        if "metrics" in sections and "evaluation_metrics" in info:
            fprint("\n📊 EVALUATION METRICS")
            fprint("-" * 70)
            metrics = info["evaluation_metrics"]
            if "primary" in metrics:
                fprint(f"  Primary Metric: {metrics['primary']}")
            if "description" in metrics:
                fprint(f"  Description: {metrics['description']}")
            if "secondary" in metrics:
                fprint(f"  Secondary Metrics: {', '.join(metrics['secondary'])}")

        # Citation
        if "citation" in sections and "citation" in info:
            fprint("\n📚 CITATION")
            fprint("-" * 70)
            fprint(info["citation"])

        # Additional Notes
        if "notes" in info and sections is None:
            fprint("\n📝 NOTES")
            fprint("-" * 70)
            for i, note in enumerate(info["notes"], 1):
                fprint(f"  {i}. {note}")

        fprint("\n" + "=" * 70 + "\n")

        if self.dataset_info is None:
            return None

        if key is None:
            return self.dataset_info

        # Support dot notation for nested keys
        keys = key.split(".")
        value = self.dataset_info
        for k in keys:
            if isinstance(value, dict) and k in value:
                value = value[k]
            else:
                return None

        return value

    def load_data_source(self, data_source, **kwargs):
        """
        Loads data from a file or list of files.

        Args:
            data_source (str or list): Path to the data file or a list of paths.
            **kwargs: Additional keyword arguments, e.g., `max_examples`.

        Returns:
            list: A list of examples.
        """
        examples = []
        max_examples = kwargs.get("max_examples", None)
        columns = kwargs.get("select_columns", None)
        if not isinstance(data_source, list):
            data_source = [data_source]

        for data_source in data_source:
            _examples = []

            if data_source.endswith(".csv"):
                import pandas as pd

                df = pd.read_csv(data_source)
                for i in range(len(df)):
                    _examples.append(df.iloc[i].to_dict())
            elif data_source.endswith(".json") or data_source.endswith(".jsonl"):
                import json

                try:
                    with open(data_source, "r", encoding="utf8") as f:
                        _examples = json.load(f)
                except:
                    with open(data_source, "r", encoding="utf8") as f:
                        lines = f.readlines()  # Assume the data is a list of examples
                    for i in range(len(lines)):
                        lines[i] = json.loads(lines[i])
                    for line in lines:
                        _examples.append(line)
            elif data_source.endswith(".parquet"):
                import pandas as pd

                df = pd.read_parquet(data_source)
                for i in range(len(df)):
                    _examples.append(df.iloc[i].to_dict())
            elif data_source.endswith(".npy") or data_source.endswith(".npz"):
                import numpy as np

                if data_source.endswith(".npy"):
                    data = np.load(data_source, allow_pickle=True)
                    if isinstance(data, np.ndarray):
                        for item in data:
                            _examples.append(
                                {
                                    "sequence": item["sequence"],
                                    "label": item.get("label", None),
                                }
                            )
                    else:
                        raise ValueError(
                            "Unexpected data format in .npy file, expected an array of dictionaries. e.g.,"
                            " [{'sequence': 'ATCG', 'label': 1}, ...]"
                        )
                elif data_source.endswith(".npz"):
                    data = np.load(data_source, allow_pickle=True)
                    for key in data.files:
                        item = data[key]
                        if isinstance(item, np.ndarray):
                            for sub_item in item:
                                _examples.append(
                                    {
                                        "sequence": sub_item["sequence"],
                                        "label": sub_item.get("label", None),
                                    }
                                )
                        else:
                            raise ValueError(
                                "Unexpected data format in .npz file, expected an array of dictionaries. e.g.,"
                                " [{'sequence': 'ATCG', 'label': 1}, ...]"
                            )
            elif data_source.endswith(
                (".fasta", ".fa", ".fna", ".ffn", ".faa", ".frn")
            ):
                try:
                    from Bio import SeqIO
                except ImportError:
                    raise ImportError(
                        "Biopython is required for FASTA parsing. Please install with 'pip install biopython'."
                    )
                for record in SeqIO.parse(data_source, "fasta"):
                    _examples.append(
                        {
                            "id": record.id,
                            "sequence": str(record.seq),
                            "description": record.description,
                        }
                    )
            elif data_source.endswith((".fastq", ".fq")):
                try:
                    from Bio import SeqIO
                except ImportError:
                    raise ImportError(
                        "Biopython is required for FASTQ parsing. Please install with 'pip install biopython'."
                    )
                for record in SeqIO.parse(data_source, "fastq"):
                    _examples.append(
                        {
                            "id": record.id,
                            "sequence": str(record.seq),
                            "quality": record.letter_annotations.get(
                                "phred_quality", []
                            ),
                        }
                    )
            elif data_source.endswith(".bed"):
                import pandas as pd

                df = pd.read_csv(data_source, sep="\t", comment="#")
                for i in range(len(df)):
                    _examples.append(df.iloc[i].to_dict())

            else:
                raise Exception("Unknown file format.")

            if columns is not None:
                fprint(f"Selecting columns: {columns}")
                filtered_examples = []
                for ex in _examples:
                    filtered_ex = {col: ex[col] for col in columns if col in ex}
                    filtered_examples.append(filtered_ex)
                _examples = filtered_examples

            examples.extend(_examples)
            del _examples

            fprint(
                f"Reading from {data_source}, Loaded {len(examples)} examples so far..."
            )

        if self.shuffle is True:
            fprint("Detected shuffle=True, shuffling the examples...")
            random.shuffle(examples)

        if max_examples is not None:
            fprint(f"Detected max_examples={max_examples}, truncating the examples...")
            examples = examples[:max_examples]

        self.examples = examples
        return self.examples

    def prepare_input(self, instance, **kwargs):
        """
        Prepares a single data instance for the model. Must be implemented by subclasses.

        Args:
            instance (dict): A single data instance (e.g., a dictionary).
            **kwargs: Additional keyword arguments for tokenization.

        Returns:
            dict: A dictionary of tokenized inputs.
        """
        raise NotImplementedError(
            "The prepare_input() function should be implemented for your dataset."
        )

    @staticmethod
    def _download_dataset_from_hub(dataset_name, local_dir=None):
        """
        Downloads and extracts datasets from OmniGenBench Hub powered by HuggingFace.

        This method supports downloading datasets from the OmniGenBench Hub on HuggingFace.

        Args:
            dataset_name (str): Name of the dataset to download.
            local_dir (str, optional): Directory to save the dataset. If None, saves to default location.
        """
        if local_dir is None:
            local_dir = os.path.join(
                os.getcwd(), f"__OMNIGENOME_DATA__/datasets/{dataset_name}"
            )
        else:
            local_dir = os.path.abspath(local_dir)

        url_to_download = f"https://huggingface.co/datasets/yangheng/OmniGenBench_Hub/resolve/main/{dataset_name}.zip"
        zip_path = os.path.join(local_dir, f"{dataset_name}.zip")

        if not os.path.exists(local_dir):
            if not os.path.exists(local_dir):
                os.makedirs(local_dir, exist_ok=True)

            fprint(f"Downloading dataset from {url_to_download}...")
            response = requests.get(url_to_download, stream=True)
            response.raise_for_status()

            with open(zip_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
            fprint(f"Downloaded {zip_path}")

        # Unzip the dataset
        if os.path.exists(zip_path):
            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                zip_ref.extractall(local_dir)
            os.remove(zip_path)
        else:
            fprint(
                f"Dataset already downloaded and extracted at {local_dir}."
                f"If you want to re-download, please delete the existing directory."
            )

    @staticmethod
    def _download_dataset_from_huggingface(dataset_name, local_dir=None):
        """
        Downloads and extracts datasets from OmniGenBench Hub on powered by HuggingFace.

        .. deprecated::
            Use `_download_dataset_from_huggingface` instead.
        """
        warnings.warn(
            "_download_dataset_from_huggingface() is deprecated. "
            "Use _download_dataset_from_huggingface() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return OmniDataset._download_dataset_from_hub(dataset_name, local_dir)

    def _preprocessing(self):
        """
        Performs preprocessing on the loaded examples.
        This method standardizes the 'sequence' field and adds secondary structure
        information if `structure_in` is True.
        """
        for idx, ex in enumerate(self.examples):
            if (
                "seq" in self.examples[idx]
            ):  # For the RNA or DNA stored in the "seq" field
                self.examples[idx]["sequence"] = self.examples[idx]["seq"]
                del self.examples[idx]["seq"]
            if (
                "text" in self.examples[idx]
            ):  # For the RNA or DNA stored in the "text" field
                self.examples[idx]["sequence"] = self.examples[idx]["text"]
                del self.examples[idx]["text"]

            if "sequence" not in self.examples[idx]:
                warnings.warn("The 'sequence' field is missing in the raw dataset.")
        if "sequence" in self.examples[0]:
            sequences = [ex["sequence"] for ex in self.examples]
            if self.structure_in:
                structures = self.rna2structure.fold(sequences)
                for idx, (sequence, structure) in enumerate(zip(sequences, structures)):
                    self.examples[idx][
                        "sequence"
                    ] = f"{sequence}{self.tokenizer.eos_token}{structure}"

    def _postprocessing(self):
        """
        Performs postprocessing on the tokenized data.
        This method standardizes the 'labels' field and prints the label distribution
        for classification tasks.
        """
        for idx, ex in enumerate(self.data):
            if "label" in self.data[idx]:
                self.data[idx]["labels"] = self.data[idx]["label"]
                # del self.data[idx]["label"]
            # assert (
            #         "labels" in self.data[idx]
            # ), "The 'labels' field is required in the tokenized dataset."

            if "labels" not in self.data[idx] or self.data[idx]["labels"] is None:
                self.data[idx]["labels"] = torch.tensor([-100])

        if self.data[0]["labels"].dim() == 0:
            self.print_label_distribution()

    def __len__(self):
        """
        Returns the number of samples in the dataset.

        Returns:
            int: The number of samples in the dataset.
        """
        return len(self.data)

    def __getitem__(self, idx):
        """
        Returns a single data sample at the given index.

        Args:
            idx (int): The index of the sample.

        Returns:
            OmniDict: An `OmniDict` containing the data sample.
        """
        # convert the data item to a omnigenbench dict
        return OmniDict(self.data[idx])

    def sample(self, n=1) -> "OmniDataset":
        """
        Returns a random sample of n items from the dataset.

        Args:
            n (int): The number of samples to return.

        Returns:
            OmniDataset: A OmniDataset of data samples.
        """
        # Ensure n doesn't exceed dataset size
        n = min(n, len(self.data))

        # Randomly sample indices
        sampled_indices = random.sample(range(len(self.data)), n)

        # Create a new OmniDataset instance with None dataset_name_or_path to skip loading
        sampled_dataset = self.__class__(
            dataset_name_or_path=None,
            tokenizer=self.tokenizer,
            max_length=self.max_length,
            label2id=self.label2id,
            shuffle=False,  # Don't shuffle sampled data
            structure_in=self.structure_in,
            drop_long_seq=self.drop_long_seq,
            force_padding=self.force_padding,
        )

        # Copy sampled data
        sampled_dataset.data = [self.data[i] for i in sampled_indices]
        sampled_dataset.examples = (
            [self.examples[i] for i in sampled_indices] if self.examples else []
        )

        # Copy metadata and other attributes
        sampled_dataset.metadata = self.metadata.copy()
        if hasattr(self, "id2label"):
            sampled_dataset.id2label = self.id2label
        if hasattr(self, "rna2structure"):
            sampled_dataset.rna2structure = self.rna2structure

        return sampled_dataset

    def get_column(self, column_name):
        """
        Returns all values for a specific column in the dataset.

        Args:
            column_name (str): The name of the column.

        Returns:
            list: A list of values from the specified column.
        """
        return [data_item[column_name] for data_item in self.data]

    def get_labels(self):
        """
        Returns the set of unique labels in the dataset.

        Returns:
            set: The set of unique labels.
        """
        return set(self.get_column("labels"))

    def get_inputs_length(self):
        """
        Calculates and returns statistics about sequence and label lengths.

        Returns:
            dict: A dictionary with length statistics (min, max, avg).
        """
        if hasattr(self.tokenizer, "pad_token_id"):
            pad_token_id = self.tokenizer.pad_token_id
        else:
            pad_token_id = self.tokenizer.base_tokenizer.pad_token_id
        length = {}
        all_seq_lengths = [
            torch.sum(data_item["input_ids"] != pad_token_id) for data_item in self.data
        ]
        all_label_lengths = [
            data_item["labels"].shape[0] if data_item["labels"].shape else 1
            for data_item in self.data
        ]
        length["avg_seq_len"] = np.mean(all_seq_lengths)
        length["max_seq_len"] = np.max(all_seq_lengths)
        length["min_seq_len"] = np.min(all_seq_lengths)
        length["avg_label_len"] = np.mean(all_label_lengths)
        length["max_label_len"] = np.max(all_label_lengths)
        length["min_label_len"] = np.min(all_label_lengths)
        return length

    def _max_labels_length(self):
        """
        Returns the maximum length of labels in the dataset.

        Returns:
            int: The maximum length of labels.
        """
        if self.data[0]["labels"].dim() > 0:
            return max([len(ex["labels"]) for ex in self.data])
        else:
            return 1

    def __iter__(self):
        """
        Returns an iterator over the dataset.

        Returns:
            iterator: An iterator over the dataset.
        """
        for data_item in self.data:
            yield OmniDict(data_item)



# -*- coding: utf-8 -*-
# file: rna_design.py
# time: 19:06 05/02/2025
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# Homepage: https://yangheng95.github.io
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2025. All Rights Reserved.
import json
import argparse
from pathlib import Path
from omnigenbench import OmniModelForRNADesign
from ..base import BaseCommand


class RNADesignCommand(BaseCommand):
    """
    This class provides a CLI interface for designing RNA sequences that fold into
    specific secondary structures. It uses genetic algorithms with customizable
    parameters to optimize sequence design for target structures.

    The design process involves:

    1. Loading a pre-trained RNA design model
    2. Running genetic algorithm optimization
    3. Generating sequences that match the target structure
    4. Saving results to file (optional)

    Attributes:
        model_path (str): Path to the pre-trained RNA design model
        structure (str): Target RNA secondary structure in dot-bracket notation
        mutation_ratio (float): Genetic algorithm mutation rate
        num_population (int): Population size for genetic algorithm
        num_generation (int): Number of generations for evolution

    Example:
        >>> # Basic RNA design
        >>> python -m omnigenbench.cli design --structure "(((...)))"
        >>> # Design with custom parameters
        >>> python -m omnigenbench.cli design
        ...     --structure "(((...)))"
        ...     --model-path "yangheng/OmniGenome-186M"
        ...     --mutation-ratio 0.3
        ...     --num-population 200
        ...     --num-generation 150
        ...     --output "results.json"
    """

    @classmethod
    def register_command(cls, subparsers):
        """
        This method sets up the command-line interface for RNA sequence design,
        including all necessary arguments and their descriptions.

        Args:
            subparsers: The subparsers object from argparse to add the command to

        Example:
            >>> parser = argparse.ArgumentParser()
            >>> subparsers = parser.add_subparsers()
            >>> RNADesignCommand.register_command(subparsers)
        """
        parser: argparse.ArgumentParser = subparsers.add_parser(
            "design",
            help="RNA Sequence Design based on Secondary Structure, Using Genetic Algorithm by OmniGenome",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        parser.add_argument(
            "--structure",
            required=True,
            help="The target RNA structure in dot-bracket notation (e.g., '(((...)))')",
        )
        parser.add_argument(
            "--model-path",
            default="yangheng/OmniGenome-186M",
            help="Model path to the pre-trained model (default: yangheng/OmniGenome-186M)",
        )
        parser.add_argument(
            "--mutation-ratio",
            type=float,
            default=0.5,
            help="Mutation ratio for genetic algorithm (0.0-1.0, default: 0.5)",
        )
        parser.add_argument(
            "--num-population",
            type=int,
            default=100,
            help="Number of individuals in population (default: 100)",
        )
        parser.add_argument(
            "--num-generation",
            type=int,
            default=100,
            help="Number of generations to evolve (default: 100)",
        )
        parser.add_argument(
            "--output", type=Path, help="Output JSON file to save results"
        )
        cls.add_common_arguments(parser)
        parser.set_defaults(func=cls.execute)

    @staticmethod
    def execute(args: argparse.Namespace):
        """
        This method runs the RNA sequence design process using genetic algorithms.
        It validates parameters, loads the model, runs the design optimization,
        and outputs or saves the results.

        Args:
            args (argparse.Namespace): Parsed command-line arguments containing
                                      design parameters and model settings

        Raises:
            ValueError: If mutation_ratio is not between 0.0 and 1.0

        Example:
            >>> args = parser.parse_args(['design', '--structure', '(((...)))'])
            >>> RNADesignCommand.execute(args)
        """
        # 参数验证逻辑
        if not 0 <= args.mutation_ratio <= 1:
            raise ValueError("--mutation-ratio should be between 0.0 and 1.0")

        # 核心业务逻辑
        model = OmniModelForRNADesign(model_path=args.model_path)
        best_sequences = model.design(
            structure=args.structure,
            mutation_ratio=args.mutation_ratio,
            num_population=args.num_population,
            num_generation=args.num_generation,
        )

        # 结果输出
        print(f"The best RNA sequences for {args.structure}:")
        for seq in best_sequences:
            print(f"- {seq}")

        # 结果保存
        if args.output:
            args.output.parent.mkdir(parents=True, exist_ok=True)
            with open(args.output, "w") as f:
                json.dump(
                    {
                        "structure": args.structure,
                        "parameters": vars(args),
                        "best_sequences": best_sequences,
                    },
                    f,
                    indent=2,
                )
            print(f"\nResults saved to {args.output}")


def register_command(subparsers):
    """
    This function is a convenience wrapper for registering the RNADesignCommand
    with the argument parser.

    Args:
        subparsers: The subparsers object from argparse to add the command to

    Example:
        >>> parser = argparse.ArgumentParser()
        >>> subparsers = parser.add_subparsers()
        >>> register_command(subparsers)
    """
    RNADesignCommand.register_command(subparsers)



# -*- coding: utf-8 -*-
# file: auto_train.py
# time: 11:54 14/04/2024
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2024. All Rights Reserved.

import os
import time
import warnings

import findfile
import torch
from metric_visualizer import MetricVisualizer
from transformers import TrainingArguments, Trainer as HFTrainer

from ...auto.config.auto_config import AutoConfig
from ...src.lora.lora_model import OmniLoraModel
from ...src.abc.abstract_tokenizer import OmniTokenizer
from ...src.misc.utils import (
    seed_everything,
    fprint,
    load_module_from_path,
    clean_temp_checkpoint,
)
from ...src.trainer.accelerate_trainer import AccelerateTrainer
from ...src.trainer.trainer import Trainer

autotrain_evaluations = "./autotrain_evaluations"


class AutoTrain:
    """
    This class provides a comprehensive framework for training genomic models
    on various datasets with minimal configuration. It handles dataset loading,
    model initialization, training configuration, and result tracking.

    AutoTrain supports various training scenarios including:

    - Single dataset training with multiple seeds
    - Different trainer backends (native, accelerate, huggingface)
    - Automatic metric visualization and result tracking
    - Configurable training parameters

    Attributes:
        dataset (str): The name or path of the dataset to use for training.
        model_name_or_path (str): The name or path of the model to train.
        tokenizer: The tokenizer to use for training.
        autocast (str): The autocast precision to use ('fp16', 'bf16', etc.).
        overwrite (bool): Whether to overwrite existing training results.
        trainer (str): The trainer to use ('native', 'accelerate', 'hf_trainer').
        mv_path (str): Path to the metric visualizer file.
        mv (MetricVisualizer): The metric visualizer instance.
    """

    def __init__(
        self,
        dataset,
        model_name_or_path,
        tokenizer=None,
        **kwargs,
    ):
        """
        Initialize the AutoTrain instance.

        Args:
            dataset (str): The name or path of the dataset to use for training.
            model_name_or_path (str): The model instance, model name or model path of the model to train.
            tokenizer: The tokenizer to use. If None, it will be loaded from the model path.
            **kwargs: Additional keyword arguments.
                - autocast (str): The autocast precision to use ('fp16', 'bf16', etc.).
                  Defaults to 'fp16'.
                - overwrite (bool): Whether to overwrite existing training results.
                  Defaults to False.
                - trainer (str): The trainer to use ('native', 'accelerate', 'hf_trainer').
                  Defaults to 'accelerate'.

        Example:
            >>> # Initialize with a dataset and model
            >>> trainer = AutoTrain("dataset_name_or_path", "model_name")

            >>> # Initialize with custom settings
            >>> trainer = AutoTrain("dataset_name_or_path", "model_name",
            ...                     autocast="bf16", trainer="accelerate")
        """
        self.dataset = dataset.rstrip("/")
        self.autocast = kwargs.pop("autocast", "fp16")
        self.overwrite = kwargs.pop("overwrite", False)
        self.trainer = kwargs.pop("trainer", "accelerate")

        self.model_name_or_path = model_name_or_path
        self.tokenizer = tokenizer
        if isinstance(self.model_name_or_path, str):
            self.model_name_or_path = self.model_name_or_path.rstrip("/")
            self.model_name = self.model_name_or_path.split("/")[-1]
        else:
            self.model_name = self.model_name_or_path.__class__.__name__
        if isinstance(tokenizer, str):
            self.tokenizer = tokenizer.rstrip("/")
        os.makedirs(autotrain_evaluations, exist_ok=True)
        time_str = time.strftime("%Y%m%d_%H%M%S", time.localtime())
        mv_name = f"{dataset}-{self.model_name}"
        self.mv_path = f"{autotrain_evaluations}/{mv_name}-{time_str}.mv"

        mv_paths = findfile.find_files(
            autotrain_evaluations,
            [dataset, self.model_name, ".mv"],
        )
        if mv_paths and not self.overwrite:
            self.mv = MetricVisualizer.load(mv_paths[-1])
            self.mv.summary(round=4)
        else:
            self.mv = MetricVisualizer(self.mv_path)
        self.train_info()

    def train_info(self):
        """
        Print and return information about the current training setup.

        Returns:
            str: A string containing training setup information.

        Example:
            >>> info = trainer.train_info()
            >>> print(info)
        """
        info = f"Dataset Root: {self.dataset}\n"
        info += f"Model Name or Path: {self.model_name}\n"
        info += f"Tokenizer: {self.tokenizer}\n"
        info += f"Metric Visualizer Path: {self.mv_path}\n"
        fprint(info)
        return info

    def run(self, **kwargs):
        """
        This method loads the dataset configuration, initializes the model and
        tokenizer, and runs training across multiple seeds. It supports various
        training backends and automatic result tracking.

        Args:
            **kwargs: Additional keyword arguments that will override the default
                     parameters in the dataset configuration.

        Example:
            >>> # Run training with default settings
            >>> trainer.run()
            >>> # Run with custom parameters
            >>> trainer.run(learning_rate=1e-4, batch_size=16)
        """

        clean_temp_checkpoint(1)  # clean temp checkpoint older than 1 day

        _kwargs = kwargs.copy()

        train_config_path = findfile.find_file(
            self.dataset,
            f"{self.dataset}.config".split("."),
        )
        config = load_module_from_path("config", train_config_path)
        train_config = None
        for attr_name in dir(config):
            attr = getattr(config, attr_name)
            if isinstance(attr, AutoConfig):  # Check if it is an instance of AutoConfig
                train_config = attr
        if train_config is None:
            raise ValueError(
                f"Could not find AutoConfig instance in {train_config_path}"
            )
        fprint(f"Loaded config for {self.dataset} from {train_config_path}")
        fprint(train_config.args)

        # Init Tokenizer and Model
        if not self.tokenizer:
            tokenizer = OmniTokenizer.from_pretrained(
                self.model_name_or_path, trust_remote_code=True
            )
        else:
            tokenizer = self.tokenizer

        for key, value in _kwargs.items():
            if key in train_config:
                fprint("Override", key, "with", value, "according to the input kwargs")
                train_config.update({key: value})

            else:
                warnings.warn(
                    f"kwarg: {key} not found in train_config while setting {key} = {value}"
                )
                train_config.update({key: value})

        for key, value in train_config.items():
            if key in train_config and key in _kwargs:
                _kwargs.pop(key)
        fprint(
            f"Autotrain Config for {self.dataset}:",
            "\n".join([f"{k}: {v}" for k, v in train_config.items()]),
        )

        if not isinstance(train_config["seeds"], list):
            train_config["seeds"] = [train_config["seeds"]]

        random_seeds = train_config["seeds"]
        for seed in random_seeds:
            batch_size = (
                train_config["batch_size"] if "batch_size" in train_config else 8
            )

            record_name = f"{os.path.basename(self.dataset)}-{self.model_name}".split(
                "/"
            )[-1]

            # check if the record exists
            if record_name in self.mv.transpose() and len(
                list(self.mv.transpose()[record_name].values())[0]
            ) >= len(random_seeds):
                continue

            seed_everything(seed)
            if self.model_name_or_path:
                model_cls = train_config["model_cls"]
                model = model_cls(
                    self.model_name_or_path,
                    tokenizer=tokenizer,
                    label2id=train_config.label2id,
                    num_labels=train_config["num_labels"],
                    trust_remote_code=True,
                    ignore_mismatched_sizes=True,
                )
            else:
                raise ValueError(
                    "model_name_or_path is not specified. Please provide a valid model name or path."
                )

            if kwargs.get("lora_config", None) is not None:
                fprint("Applying LoRA to the model with config:", kwargs["lora_config"])
                model = OmniLoraModel(model, **kwargs.get("lora_config", {}))

            # Init Trainer
            dataset_cls = train_config["dataset_cls"]

            if hasattr(model.config, "max_position_embeddings"):
                max_length = min(
                    train_config["max_length"],
                    model.config.max_position_embeddings,
                )
            else:
                max_length = train_config["max_length"]

            train_set = dataset_cls(
                dataset_name_or_path=train_config["train_file"],
                tokenizer=tokenizer,
                label2id=train_config["label2id"],
                max_length=max_length,
                structure_in=train_config.get("structure_in", False),
                max_examples=train_config.get("max_examples", None),
                shuffle=train_config.get("shuffle", True),
                drop_long_seq=train_config.get("drop_long_seq", False),
                **_kwargs,
            )
            test_set = dataset_cls(
                dataset_name_or_path=train_config["test_file"],
                tokenizer=tokenizer,
                label2id=train_config["label2id"],
                max_length=max_length,
                structure_in=train_config.get("structure_in", False),
                max_examples=train_config.get("max_examples", None),
                shuffle=False,
                drop_long_seq=train_config.get("drop_long_seq", False),
                **_kwargs,
            )
            valid_set = dataset_cls(
                dataset_name_or_path=train_config["valid_file"],
                tokenizer=tokenizer,
                label2id=train_config["label2id"],
                max_length=max_length,
                structure_in=train_config.get("structure_in", False),
                max_examples=train_config.get("max_examples", None),
                shuffle=False,
                drop_long_seq=train_config.get("drop_long_seq", False),
                **_kwargs,
            )

            if self.trainer == "hf_trainer":
                # Set up HuggingFace Trainer
                hf_kwargs = {
                    k: v
                    for k, v in kwargs.items()
                    if hasattr(TrainingArguments, k) and k != "output_dir"
                }
                training_args = TrainingArguments(
                    output_dir=f"./autotrain_evaluations/{self.model_name}",
                    num_train_epochs=hf_kwargs.pop(
                        "num_train_epochs", train_config["epochs"]
                    ),
                    per_device_train_batch_size=hf_kwargs.pop("batch_size", batch_size),
                    per_device_eval_batch_size=hf_kwargs.pop("batch_size", batch_size),
                    gradient_accumulation_steps=hf_kwargs.pop(
                        "gradient_accumulation_steps", 1
                    ),
                    learning_rate=hf_kwargs.pop("learning_rate", 2e-5),
                    weight_decay=hf_kwargs.pop("weight_decay", 0),
                    eval_strategy=hf_kwargs.pop("eval_strategy", "epoch"),
                    save_strategy=hf_kwargs.pop("save_strategy", "epoch"),
                    fp16=hf_kwargs.pop("fp16", True),
                    remove_unused_columns=False,
                    label_names=["labels"],
                    **hf_kwargs,
                )

                valid_set = valid_set if len(valid_set) else test_set

                if len(train_config["compute_metrics"]) > 1:
                    fprint(
                        "Multiple metrics not supported by HFTrainer, using the first one metric only."
                    )
                trainer = HFTrainer(
                    model=model,
                    args=training_args,
                    train_dataset=train_set,
                    eval_dataset=valid_set,
                    compute_metrics=(
                        train_config["compute_metrics"][0]
                        if isinstance(train_config["compute_metrics"], list)
                        else train_config["compute_metrics"]
                    ),
                )

                # Train and evaluate
                eval_result = trainer.evaluate(
                    valid_set if len(valid_set) else test_set
                )
                print(eval_result)
                train_result = trainer.train()
                eval_result = trainer.evaluate()
                test_result = trainer.evaluate(test_set if len(test_set) else valid_set)

                metrics = {
                    "train": train_result.metrics,
                    "eval": eval_result,
                    "test": test_result,
                }
                fprint(metrics)
            else:
                optimizer = torch.optim.AdamW(
                    filter(lambda p: p.requires_grad, model.parameters()),
                    lr=(
                        train_config["learning_rate"]
                        if "learning_rate" in train_config
                        else 2e-5
                    ),
                    weight_decay=(
                        train_config["weight_decay"]
                        if "weight_decay" in train_config
                        else 0
                    ),
                )
                if self.trainer == "accelerate":
                    trainer_cls = AccelerateTrainer
                else:
                    trainer_cls = Trainer
                fprint(f"Using Trainer: {trainer_cls}")
                trainer = trainer_cls(
                    model=model,
                    train_dataset=train_set,
                    eval_dataset=valid_set,
                    test_dataset=test_set,
                    batch_size=batch_size,
                    patience=(
                        train_config["patience"] if "patience" in train_config else 3
                    ),
                    epochs=train_config["epochs"],
                    gradient_accumulation_steps=train_config.get(
                        "gradient_accumulation_steps", 1
                    ),
                    optimizer=optimizer,
                    loss_fn=(
                        train_config["loss_fn"] if "loss_fn" in train_config else None
                    ),
                    compute_metrics=train_config["compute_metrics"],
                    seed=seed,
                    autocast=self.autocast,
                    **_kwargs,
                )
                metrics = trainer.train()
                print(_kwargs)
                if _kwargs.get("save_model", True):
                    fprint(
                        f"Saving model to {autotrain_evaluations}/{self.dataset}/{self.model_name}"
                    )
                    save_path = os.path.join(
                        autotrain_evaluations, self.dataset, self.model_name
                    )
                    os.makedirs(save_path, exist_ok=True)
                    trainer.save_model(save_path, overwrite=True)

                if metrics:
                    for key, value in metrics["test"][-1].items():
                        try:
                            value = float(value)
                        except:
                            pass  # ignore non-float values
                        self.mv.log(f"{record_name}", f"{key}", value)
                    # for key, value in metrics['test'][-1].items():
                    #     self.mv.log(f'{record_name}', f'test_{key}', value)
                    # for i, valid_metrics in enumerate(metrics["valid"]):
                    #     for key, value in valid_metrics.items():
                    #         self.mv.log(f'{record_name}', f'valid_epoch_{i}_{key}', value)

                    self.mv.summary(round=4)
                    self.mv.dump(self.mv_path)
                    self.mv.to_csv(self.mv_path.replace(".mv", ".csv"))
                del model, trainer, optimizer
                torch.cuda.empty_cache()



# -*- coding: utf-8 -*-
# file: regression_metric.py
# time: 12:57 09/04/2024
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2024. All Rights Reserved.


import types
import warnings

import numpy as np
import sklearn.metrics as metrics

from ..abc.abstract_metric import OmniMetric


def mcrmse(y_true, y_pred):
    """
    Compute Mean Column Root Mean Square Error (MCRMSE).

    MCRMSE is a multi-target regression metric that computes the RMSE for each target
    column and then takes the mean across all targets.

    Args:
        y_true (np.ndarray): Ground truth values with shape (n_samples, n_targets)
        y_pred (np.ndarray): Predicted values with shape (n_samples, n_targets)

    Returns:
        float: Mean Column Root Mean Square Error

    Raises:
        ValueError: If y_true and y_pred have different shapes

    Example:
        >>> y_true = np.array([[1, 2], [3, 4], [5, 6]])
        >>> y_pred = np.array([[1.1, 2.1], [2.9, 4.1], [5.2, 5.8]])
        >>> mcrmse(y_true, y_pred)
        0.1833...
    """
    if y_true.shape != y_pred.shape:
        raise ValueError("y_true and y_pred must have the same shape")
    mask = y_true != -100
    filtered_y_pred = y_pred[mask]
    filtered_y_true = y_true[mask]
    rmse_per_target = np.sqrt(np.mean((filtered_y_true - filtered_y_pred) ** 2, axis=0))
    mcrmse_value = np.mean(rmse_per_target)
    return mcrmse_value


setattr(metrics, "mcrmse", mcrmse)


class Metric(OmniMetric):
    """
    A flexible metric class that provides access to all scikit-learn metrics
    and custom metrics for evaluation.

    This class dynamically wraps scikit-learn metrics and provides a unified
    interface for computing various evaluation metrics. It handles different
    input formats including HuggingFace trainer outputs and supports
    custom metric functions.

    Attributes:
        metric_func: Custom metric function if provided
        ignore_y: Value to ignore in predictions and true values
        kwargs: Additional keyword arguments for metric computation
        metrics: Dictionary of available metrics including custom ones

    Example:
        >>> from omnigenbench import Metric
        >>> metric = Metric(ignore_y=-100)
        >>> y_true = [0, 1, 2, 0, 1]
        >>> y_pred = [0, 1, 1, 0, 1]
        >>> result = metric.accuracy(y_true, y_pred)
        >>> print(result)
        {'accuracy': 0.8}
    """

    def __init__(self, metric_func=None, ignore_y=-100, *args, **kwargs):
        """
        Initialize the Metric class.

        Args:
            metric_func (callable, optional): Custom metric function to use
            ignore_y (int, optional): Value to ignore in predictions and true values. Defaults to -100
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments for metric computation
        """
        super().__init__(metric_func, ignore_y, *args, **kwargs)
        self.kwargs = kwargs
        self.metrics = {"mcrmse": mcrmse}
        for key, value in metrics.__dict__.items():
            setattr(self, key, value)

    def __getattribute__(self, name):
        """
        Dynamically create metric computation methods.

        This method intercepts attribute access and creates wrapper functions
        for scikit-learn metrics, handling different input formats and
        preprocessing the data appropriately.

        Args:
            name (str): Name of the metric to access

        Returns:
            callable: Wrapper function for the requested metric
        """
        # Get the metric function
        metric_func = getattr(metrics, name, None)

        if metric_func and isinstance(metric_func, types.FunctionType):
            setattr(self, "compute", metric_func)
            # If the metric function exists, return a wrapper function

            def wrapper(y_true=None, y_score=None, *args, **kwargs):
                """
                Compute the metric, based on the true and predicted values.

                This wrapper handles different input formats including HuggingFace
                trainer outputs and performs necessary preprocessing.

                Args:
                    y_true: The true values or HuggingFace EvalPrediction object
                    y_score: The predicted values
                    ignore_y: The value to ignore in the predictions and true values in corresponding positions
                    *args: Additional positional arguments for the metric
                    **kwargs: Additional keyword arguments for the metric

                Returns:
                    dict: Dictionary containing the metric name and computed value

                Raises:
                    ValueError: If neither y_true nor y_score is provided
                """
                # This is an ugly method to handle the case when the predictions are in the form of a tuple
                # for huggingface trainers
                if y_true is not None and y_score is None:
                    if hasattr(y_true, "predictions"):
                        y_score = y_true.predictions
                    if hasattr(y_true, "label_ids"):
                        y_true = y_true.label_ids
                    if hasattr(y_true, "labels"):
                        y_true = y_true.labels
                    if len(y_score[0][1]) == np.max(y_true) + 1:
                        y_score = y_score[0]
                    else:
                        y_score = y_score[1]
                    y_score = np.argmax(y_score, axis=1)
                elif y_true is not None and y_score is not None:
                    pass  # y_true and y_score are provided
                else:
                    raise ValueError(
                        "Please provide the true and predicted values or a dictionary with 'y_true' and 'y_score'."
                    )

                y_true, y_score = Metric.flatten(y_true, y_score)
                y_true_mask_idx = np.where(y_true != self.ignore_y)
                if self.ignore_y is not None:
                    y_true = y_true[y_true_mask_idx]
                    try:
                        y_score = y_score[y_true_mask_idx]
                    except Exception as e:
                        warnings.warn(str(e))
                kwargs.update(self.kwargs)

                return {name: self.compute(y_true, y_score, *args, **kwargs)}

            return wrapper
        else:
            return super().__getattribute__(name)

    def compute(self, y_true, y_score, *args, **kwargs):
        """
        Compute the metric, based on the true and predicted values.

        Args:
            y_true: The true values
            y_score: The predicted values
            *args: Additional positional arguments for the metric
            **kwargs: Additional keyword arguments for the metric

        Returns:
            The computed metric value

        Raises:
            NotImplementedError: If no metric function is provided and compute is not implemented
        """
        if self.metric_func is not None:
            kwargs.update(self.kwargs)
            return self.metric_func(y_true, y_score, *args, **kwargs)

        else:
            raise NotImplementedError(
                "Method compute() is not implemented in the child class."
            )



# -*- coding: utf-8 -*-
# file: omnigenbench_model.py
# time: 18:36 06/04/2024
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2024. All Rights Reserved.
import json
import os
import shutil
import time
import warnings
import inspect
from importlib import import_module

import dill
import findfile
import torch
from transformers import AutoModel, AutoConfig, AutoTokenizer, BatchEncoding

from ..misc.utils import fprint, env_meta_info

warnings.filterwarnings("once")


def count_parameters(model):
    """
    This function iterates through all parameters of a PyTorch model and counts
    only those that require gradients (i.e., trainable parameters).

    Args:
        model (torch.nn.Module): A PyTorch model.

    Returns:
        int: The total number of trainable parameters.

    Example:
        >>> model = OmniModelForSequenceClassification(config, tokenizer)
        >>> num_params = count_parameters(model)
        >>> print(f"Model has {num_params} trainable parameters")
    """
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


class OmniModel(torch.nn.Module):
    """
    This class provides a unified interface for all genomic models in the OmniGenome
    framework. It handles model initialization, forward passes, loss computation,
    prediction, inference, and model persistence.
    """

    def __init__(self, config_or_model, tokenizer, *args, **kwargs):
        """
        Initializes the model.

        This method handles different types of model initialization:
        - From a pre-trained model path (string)
        - From a PyTorch model instance
        - From a configuration object

        Args:
            config_or_model: A model configuration, a pre-trained model path (str),
                           or a `torch.nn.Module` instance.
            tokenizer: The tokenizer associated with the model.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
                - label2id (dict): Mapping from class labels to IDs.
                - num_labels (int): The number of labels.
                - trust_remote_code (bool): Whether to trust remote code when loading
                  from Hugging Face Hub. Defaults to True.
                - ignore_mismatched_sizes (bool): Whether to ignore size mismatches
                  when loading pre-trained weights. Defaults to False.
                - dropout (float): Dropout rate. Defaults to 0.0.
                - dataset_class: The dataset class used for data preprocessing.
                  This allows the model to use dataset's prepare_input method
                  during inference. Defaults to None.

        Raises:
            ValueError: If config_or_model is not a valid type or if required
                      configuration is missing.
            RuntimeError: If the hidden size cannot be determined from the config.

        Example:
            >>> # Initialize from a pre-trained model
            >>> model = OmniModelForSequenceClassification("model_path", tokenizer)

            >>> # Initialize from a configuration
            >>> config = AutoConfig.from_pretrained("model_path")
            >>> model = OmniModelForSequenceClassification(config, tokenizer)
        """
        self.loss_fn = None

        label2id = kwargs.pop("label2id", None)
        trust_remote_code = kwargs.pop("trust_remote_code", True)
        num_labels = kwargs.pop("num_labels", None)
        ignore_mismatched_sizes = kwargs.pop("ignore_mismatched_sizes", False)
        dataset_class = kwargs.pop("dataset_class", None)

        if label2id is not None and num_labels is None:
            num_labels = len(label2id)
        elif num_labels is not None and label2id is None:
            label2id = {str(i): i for i in range(num_labels)}
        elif label2id is None and num_labels is None:
            raise ValueError(
                "Either label2id or num_labels must be provided to initialize the model."
            )
        else:
            if len(label2id) != num_labels:
                raise ValueError(
                    "The length of label2id does not match num_labels. "
                    f"Expected {num_labels}, but got {len(label2id)}."
                )

        # do not change the order of the following lines
        super().__init__(*args, **kwargs)

        if isinstance(config_or_model, str):
            config = AutoConfig.from_pretrained(
                config_or_model,
                num_labels=num_labels,
                label2id=label2id,
                trust_remote_code=trust_remote_code,
            )
            # Load the model from either `architectures` or `auto_map`
            if hasattr(config, "auto_map") and config.auto_map:
                architectures = list(set(config.auto_map.keys()) - set(["AutoConfig"]))
                if architectures:
                    model_cls_name = (
                        "AutoModel"
                        if "AutoModel" in architectures
                        else architectures[-1]
                    )
                    if "multimolecule" in config_or_model.__repr__().lower():
                        model_cls = getattr(
                            import_module(f"multimolecule"), model_cls_name
                        )
                    else:
                        model_cls = getattr(
                            import_module(f"transformers"), model_cls_name
                        )

                    model = model_cls.from_pretrained(
                        config_or_model,
                        config=config,
                        trust_remote_code=trust_remote_code,
                        ignore_mismatched_sizes=ignore_mismatched_sizes,
                    ).base_model
                else:
                    raise ValueError(
                        f"The model cannot be instantiated from {config_or_model}. "
                        f"Please check the model configuration contains the architectures or auto_map."
                    )
            elif hasattr(config, "architectures") and config.architectures:
                model_cls_name = (
                    AutoModel
                    if "AutoModel" in config.architectures
                    else config.architectures[-1]
                )
                if "multimolecule" in config_or_model.__repr__().lower():
                    model_cls = getattr(import_module(f"multimolecule"), model_cls_name)
                else:
                    model_cls = getattr(import_module(f"transformers"), model_cls_name)
                model = model_cls.from_pretrained(
                    config_or_model,
                    config=config,
                    trust_remote_code=trust_remote_code,
                    ignore_mismatched_sizes=ignore_mismatched_sizes,
                ).base_model
            else:
                raise ValueError(
                    "Neither `architectures` nor `auto_map` is defined in the config."
                )
            self.model = model
            self.model.config = config
            del model_cls
        elif isinstance(config_or_model, torch.nn.Module):
            self.model = config_or_model
            self.model.config.num_labels = (
                num_labels if len(label2id) == num_labels else len(label2id)
            )
            self.model.config.label2id = label2id
        elif isinstance(config_or_model, AutoConfig):
            config = config_or_model
            config.num_labels = (
                num_labels if len(label2id) == num_labels else len(label2id)
            )
            config.label2id = label2id
            self.model = AutoModel.from_config(config)
            self.model.config = config
        else:
            raise ValueError(
                "The config_or_model should be either a string, a torch.nn.Module or a AutoConfig object."
            )

        # Update the config
        self.config = self.model.config
        if isinstance(label2id, dict):
            self.config.label2id = label2id
            self.config.id2label = {v: k for k, v in label2id.items()}
        if (
            not hasattr(self.config, "num_labels")
            or len(self.config.id2label) != self.config.num_labels
        ):
            fprint(
                "Warning: The number of labels in the config is not equal to the number of labels in the label2id dictionary. "
            )
            fprint(
                "Please check the label2id dictionary and the num_labels parameter in the config."
            )
            self.config.num_labels = len(self.config.id2label)

        assert (
            len(self.config.label2id) == num_labels
        ), f"Expected {num_labels} labels, but got {len(self.config.label2id)} in label2id dictionary."

        # The metadata of the model
        self.metadata = env_meta_info()
        self.metadata["model_cls"] = self.__class__.__name__

        # Store dataset class for data preprocessing during inference
        if dataset_class is not None:
            self.dataset_class = dataset_class
            self.metadata["dataset_cls"] = dataset_class.__name__
            self.metadata["dataset_module"] = dataset_class.__module__
        elif hasattr(self, "dataset_class"):
            dataset_class = self.dataset_class
            self.metadata["dataset_cls"] = dataset_class.__name__
            self.metadata["dataset_module"] = dataset_class.__module__
        else:
            self.dataset_class = None
            self.metadata["dataset_cls"] = None
            self.metadata["dataset_module"] = None
            fprint(
                "Warning: No dataset_class is provided for the model, please set 'dataset_class=...' "
                "when initializing the model if you want to use the dataset's prepare_input method during inference."
            )

        # The config of the model
        if hasattr(self.config, "n_embd") and self.config.n_embd:
            self.config.hidden_size = self.config.n_embd
        elif hasattr(self.config, "d_model") and self.config.d_model:
            self.config.hidden_size = self.config.d_model
        elif hasattr(self.config, "hidden_size") and self.config.hidden_size:
            self.config.hidden_size = self.config.hidden_size
        else:
            raise RuntimeError(
                "The hidden size of the model is not found in the config."
            )

        # The tokenizer of the model
        self.tokenizer = tokenizer
        self.metadata["tokenizer_cls"] = self.tokenizer.__class__.__name__
        if hasattr(self.tokenizer, "base_tokenizer"):
            self.pad_token_id = self.tokenizer.base_tokenizer.pad_token_id
        else:
            self.pad_token_id = self.tokenizer.pad_token_id

        self.dropout = torch.nn.Dropout(kwargs.get("dropout", 0.0))
        self.activation = torch.nn.Tanh()

    def last_hidden_state_forward(self, **inputs):
        """
        Performs a forward pass to get the last hidden state from the base model. It also handles compatibility with different
        model architectures by mapping input parameters appropriately.

        Args:
            **inputs: The inputs to the model, compatible with the base model's
                     forward method. Typically includes 'input_ids', 'attention_mask',
                     and other model-specific parameters.

        Returns:
            torch.Tensor: The last hidden state tensor.

        Example:
            >>> inputs = {
            ...     'input_ids': torch.tensor([[1, 2, 3, 4]]),
            ...     'attention_mask': torch.tensor([[1, 1, 1, 1]])
            ... }
            >>> hidden_states = model.last_hidden_state_forward(**inputs)
        """
        model = self.model
        input_mapping = {}
        inputs["output_hidden_states"] = True

        if "strippedhyena" in model.__class__.__name__.lower():
            inputs["x"] = inputs["input_ids"]  # For compatibility with Evo models
        if isinstance(inputs, BatchEncoding) or isinstance(inputs, dict):
            # Determine the input parameter names of the model's forward method
            forward_params = inspect.signature(model.forward).parameters
            # Map the inputs to the forward method parameters
            for param in forward_params:
                if param in inputs:
                    input_mapping[param] = inputs[param]
            # 对于未在模型签名中声明的关键参数，可以给出警告或日志
            ignored_keys = set(inputs.keys()) - set(input_mapping.keys())
            if ignored_keys:
                warnings.warn(f"Warning: Ignored keys in inputs: {ignored_keys}")

            inputs = input_mapping
        elif isinstance(inputs, tuple):
            input_ids = inputs[0]
            attention_mask = inputs[1] if len(inputs) > 1 else None
            inputs = {"input_ids": input_ids, "attention_mask": attention_mask}
        elif isinstance(inputs, torch.Tensor):
            shape = inputs.shape
            try:
                if len(shape) == 3:
                    if shape[1] == 2:
                        input_ids = inputs[:, 0]
                        attention_mask = inputs[:, 1]
                    else:
                        input_ids = inputs[0]
                        attention_mask = inputs[1] if len(inputs) > 1 else None
                elif len(shape) == 2:
                    input_ids = inputs
                    attention_mask = None
                else:
                    raise ValueError(
                        f"Failed to get the input_ids and attention_mask from the inputs, got shape {shape}."
                    )
            except:
                raise ValueError(
                    f"Failed to get the input_ids and attention_mask from the inputs, got shape {shape}."
                )
            inputs = {"input_ids": input_ids, "attention_mask": attention_mask}
        else:
            raise ValueError(
                f"The inputs should be a tuple, BatchEncoding or a dictionary-like object, got {type(inputs)}."
            )

        # 执行模型
        outputs = model(**inputs)

        if not hasattr(outputs, "last_hidden_state"):
            warnings.warn(
                f"last_hidden_state not found in the outputs from the {model.__class__.__name__} model."
            )

        if hasattr(outputs, "last_hidden_state"):
            last_hidden_state = outputs.last_hidden_state
        elif isinstance(outputs, dict) and "last_hidden_state" in outputs:
            last_hidden_state = outputs["last_hidden_state"]
        elif hasattr(outputs, "hidden_states"):
            last_hidden_state = outputs.hidden_states[-1]
        elif isinstance(outputs, (list, tuple, torch.Tensor)):
            if len(outputs) <= 2:
                # For Evo models that return a tuple of (last_hidden_state, logits)
                last_hidden_state = outputs[0]
            elif len(outputs) >= 3:
                last_hidden_state = outputs[-1]
        else:
            raise ValueError(
                f"Cannot find the last hidden state in the outputs from the {model.__class__.__name__} model, "
                f"please check the model architecture."
            )

        return last_hidden_state

    def loss_function(self, logits, labels):
        """
        Calculates the loss. This method should be implemented by concrete model classes to define
        how the loss is calculated for their specific task (classification,
        regression, etc.).

        Args:
            logits (torch.Tensor): The model's output logits.
            labels (torch.Tensor): The ground truth labels.

        Returns:
            torch.Tensor: The calculated loss.

        Raises:
            NotImplementedError: If the method is not implemented by the subclass.

        Example:
            >>> # In a classification model
            >>> loss = model.loss_function(logits, labels)
        """
        raise NotImplementedError(
            "The loss_function() function should be implemented for your model."
        )

    def set_loss_fn(self, loss_function):
        """
        Sets a custom loss function for the model. The loss function should be compatible with the
        model's output format.

        Args:
            loss_function (callable): A callable loss function that takes
                                    logits and labels as arguments.

        Example:
            >>> import torch.nn as nn
            >>> model.set_loss_fn(nn.CrossEntropyLoss())
        """
        self.loss_fn = loss_function
        try:
            self.loss_fn.weight.to(self.model.device)
        except AttributeError:
            # If the loss function does not have a weight attribute, we assume it's not weighted
            pass

    def predict(self, sequence_or_inputs, **kwargs):
        """
        This method takes raw sequences or tokenized inputs and returns
        the raw model outputs (logits, hidden states, etc.) without
        post-processing. It's useful for getting the model's direct
        predictions for further processing.

        If the model has a dataset_class, this method will use the dataset's
        prepare_input method for data preprocessing, allowing for more complex
        data preparation including custom field handling.

        Args:
            sequence_or_inputs: Can be one of:
                - str: A single sequence (e.g., "ATCGATCG")
                - list: A list of sequences (e.g., ["ATCGATCG", "GCTAGCTA"])
                - dict: A dictionary with 'sequence'/'seq' and optionally 'label'/'labels'
                        and other custom fields (e.g., {"sequence": "ATCG", "label": 1})
                - BatchEncoding/dict with 'input_ids': Already tokenized inputs
            **kwargs: Additional arguments for tokenization or dataset preparation.
                     Common options include 'max_length', 'padding', 'truncation', etc.

        Returns:
            dict: A dictionary containing the raw model outputs, typically including
                 `logits`, `last_hidden_state`, and other model-specific outputs.

        Example:
            >>> # Predict on a single sequence
            >>> outputs = model.predict("ATCGATCG")
            >>>
            >>> # Predict on multiple sequences
            >>> outputs = model.predict(["ATCGATCG", "GCTAGCTA"])
            >>>
            >>> # Predict with dict input (if dataset_class is set)
            >>> outputs = model.predict({"sequence": "ATCGATCG", "label": 1})
        """
        # Please implement the predict() function for your model
        raw_outputs = self._forward_from_raw_input(sequence_or_inputs, **kwargs)
        return raw_outputs

    def inference(self, sequence_or_inputs, **kwargs):
        """
        This method takes raw sequences or tokenized inputs and returns
        processed predictions that are ready for human consumption. It
        typically includes post-processing steps like converting logits
        to class labels or probabilities.

        If the model has a dataset_class, this method will use the dataset's
        prepare_input method for data preprocessing, allowing for more complex
        data preparation including custom field handling.

        Args:
            sequence_or_inputs: Can be one of:
                - str: A single sequence (e.g., "ATCGATCG")
                - list: A list of sequences (e.g., ["ATCGATCG", "GCTAGCTA"])
                - dict: A dictionary with 'sequence'/'seq' and optionally 'label'/'labels'
                        and other custom fields (e.g., {"sequence": "ATCG", "label": 1})
                - BatchEncoding/dict with 'input_ids': Already tokenized inputs
            **kwargs: Additional arguments for tokenization, dataset preparation, or inference.
                     Common options include 'max_length', 'padding', 'truncation', etc.

        Returns:
            dict: A dictionary containing the processed predictions, typically including
                 'predictions', 'confidence', and other human-readable outputs.

        Example:
            >>> # Inference on a single sequence
            >>> results = model.inference("ATCGATCG")
            >>> print(results['predictions'])  # Class labels
            >>>
            >>> # Inference on multiple sequences
            >>> results = model.inference(["ATCGATCG", "GCTAGCTA"])
            >>>
            >>> # Inference with dict input (if dataset_class is set)
            >>> results = model.inference({"sequence": "ATCGATCG", "label": 1})
            >>> print(results['predictions'])
        """
        # Please implement the predict() function for your model
        raw_outputs = self._forward_from_raw_input(sequence_or_inputs, **kwargs)
        return raw_outputs

    def __call__(self, **inputs):
        """
        The main forward pass of the model, suitable for training loops.

        This method is the primary interface for model forward passes during
        training. It handles both tokenized inputs and raw sequences,
        calculates loss if labels are provided, and returns a comprehensive
        output dictionary.

        Args:
            **inputs: A dictionary of tokenized inputs, potentially including
                     labels. Can also handle raw sequences that will be
                     tokenized automatically.

        Returns:
            dict: A dictionary containing logits, last_hidden_state, labels,
                  and loss (if labels were provided).

        Example:
            >>> # Training forward pass
            >>> outputs = model(
            ...     input_ids=torch.tensor([[1, 2, 3, 4]]),
            ...     attention_mask=torch.tensor([[1, 1, 1, 1]]),
            ...     labels=torch.tensor([0])
            ... )
            >>> loss = outputs['loss']
        """
        # For transformer trainer integration, we need to pop the "inputs" to be a tokenized inputs object.
        # For native trainer, the inputs are already tokenized inputs object
        labels = inputs.pop("labels", None)
        inputs = inputs.pop("inputs", inputs)
        inputs["labels"] = labels
        if isinstance(inputs, dict):
            labels = inputs.get("labels", None)
            label = inputs.get("label", None)
            labels = labels if labels is not None else label
            # if labels is None:
            #     warnings.warn(
            #         "No labels are provided in the inputs, the model will not calculate the loss."
            #     )
        elif isinstance(inputs, tuple):
            labels = inputs[1]
            inputs = inputs[0]
        elif labels is not None:
            labels = labels
        outputs = self.forward(**inputs)

        if labels is not None:
            outputs["loss"] = self._calculate_loss(outputs, labels)
        else:
            outputs["loss"] = None
        return outputs

    def _calculate_loss(self, outputs, labels):
        """
        Internal method to calculate loss if not already present in outputs.

        :param outputs: The dictionary of model outputs.
        :param labels: The ground truth labels.
        :return: The calculated loss.
        """
        loss = outputs.get("loss", None)
        if loss is not None:
            return loss

        logits = outputs["logits"]
        if logits is not None or labels is not None:
            loss = self.loss_function(logits, labels)
            return loss
        else:
            raise RuntimeError(
                "The output of the forward() function should be a dictionary-like objective"
                " and have either 'loss', or 'logits' and 'labels' attribute."
            )

    # ==================== Save Helper Methods ====================

    def _save_base_files(self, path):
        """
        Copy base model configuration files to save directory.

        This method copies essential configuration files from the original model
        directory to the target save directory, excluding weight files to avoid
        duplication. Files with extensions .bin, .json, .txt, and .py are copied.

        Args:
            path (str): Target directory path where files will be copied.

        Note:
            - pytorch_model.bin and model.safetensors are excluded as they
              will be saved separately in _save_weights()
            - Only files from the original model path are copied
        """
        for file in findfile.find_files(
            self.config.name_or_path,
            or_key=["bin", "json", "txt", "py"],
            exclude_key=["pytorch_model.bin", "model.safetensors"],
            return_relative_path=False,
        ):
            shutil.copyfile(file, f"{path}/{os.path.basename(file)}")

    def _save_custom_model_class(self, path, metadata):
        """
        Save custom model class source file if it's user-defined.

        This method detects and saves the source code of custom model classes
        that are defined outside the omnigenbench/omnigenome packages. This
        enables loading models without requiring the original source code.

        Args:
            path (str): Target directory path where the custom model file will be saved.
            metadata (dict): Metadata dictionary that will be updated with custom
                           model file information.

        Side Effects:
            - Creates 'custom_model.py' in the target directory if model is custom
            - Updates metadata with 'custom_model_file' key
            - Prints confirmation message on success
            - Prints warning on failure (non-fatal)

        Note:
            - Only saves source code for models NOT in omnigenbench/omnigenome packages
            - Failures are logged but don't interrupt the save process
        """
        try:
            model_class = self.__class__
            model_source_file = inspect.getfile(model_class)
            # Check if it's a user-defined model (not from omnigenbench/omnigenome packages)
            if (
                "omnigenbench" not in model_source_file
                and "omnigenome" not in model_source_file
            ):
                custom_model_path = os.path.join(path, "custom_model.py")
                shutil.copyfile(model_source_file, custom_model_path)
                metadata["custom_model_file"] = "custom_model.py"
                fprint(f"Saved custom model class source to: {custom_model_path}")
        except (TypeError, OSError) as e:
            fprint(f"Could not save custom model source file: {e}")

    def _save_custom_dataset_class(self, path, metadata):
        """
        Save custom dataset class source file if available.

        This method saves the source code of custom dataset classes associated
        with the model. This ensures that data preprocessing logic is preserved
        and can be used during inference without requiring the original code.

        Args:
            path (str): Target directory path where the custom dataset file will be saved.
            metadata (dict): Metadata dictionary that will be updated with custom
                           dataset file information.

        Side Effects:
            - Creates 'custom_dataset.py' in the target directory if dataset is custom
            - Updates metadata with 'custom_dataset_file' and 'custom_dataset_class' keys
            - Prints confirmation message on success
            - Prints warning on failure (non-fatal)

        Note:
            - Requires model to have 'dataset_class' attribute
            - Only saves source code for datasets NOT in omnigenbench/omnigenome packages
            - Silently returns if model has no dataset_class attribute
            - Failures are logged but don't interrupt the save process
        """
        if not hasattr(self, "dataset_class"):
            return

        try:
            dataset_class = self.dataset_class
            dataset_source_file = inspect.getfile(dataset_class)
            # Check if it's a user-defined dataset (not from omnigenbench/omnigenome)
            if (
                "omnigenbench" not in dataset_source_file
                and "omnigenome" not in dataset_source_file
            ):
                custom_dataset_path = os.path.join(path, "custom_dataset.py")
                shutil.copyfile(dataset_source_file, custom_dataset_path)
                metadata["custom_dataset_file"] = "custom_dataset.py"
                metadata["custom_dataset_class"] = dataset_class.__name__
                fprint(f"Saved custom dataset class source to: {custom_dataset_path}")
        except (TypeError, OSError, AttributeError) as e:
            fprint(f"Could not save custom dataset source file: {e}")

    def _collect_metadata(self):
        """
        Collect all metadata to be saved with the model.

        This method gathers comprehensive metadata about the model, including:
        - Loss function information (class name and module)
        - Model class information (name and module)
        - Custom attributes (num_labels, num_classes, label mappings, etc.)
        - Dataset metadata (if present)

        Returns:
            dict: A dictionary containing all model metadata with the following structure:
                {
                    'model_cls': str,           # Model class name
                    'model_module': str,        # Model module path
                    'loss_fn_class': str,       # Loss function class name (optional)
                    'loss_fn_module': str,      # Loss function module path (optional)
                    'custom_attrs': dict,       # Custom model attributes (optional)
                    'dataset_metadata': dict,   # Dataset metadata (optional)
                    ... (other metadata from self.metadata)
                }

        Note:
            - Only serializable attributes (int, float, str, bool, list, dict) are saved
            - Custom attributes checked: num_labels, num_classes, threshold,
              label2idx, idx2label, tissue_names, tissue_columns
            - Base metadata is copied from self.metadata
        """
        metadata = self.metadata.copy()

        # Loss function metadata
        if self.loss_fn is not None:
            metadata["loss_fn_class"] = self.loss_fn.__class__.__name__
            metadata["loss_fn_module"] = self.loss_fn.__class__.__module__

        # Model class metadata
        model_class = self.__class__
        metadata["model_cls"] = model_class.__name__
        metadata["model_module"] = model_class.__module__

        # Custom attributes
        custom_attrs = {}
        for attr_name in [
            "num_labels",
            "num_classes",
            "threshold",
            "label2idx",
            "idx2label",
            "tissue_names",
            "tissue_columns",
        ]:
            if hasattr(self, attr_name):
                attr_value = getattr(self, attr_name)
                if isinstance(attr_value, (int, float, str, bool, list, dict)):
                    custom_attrs[attr_name] = attr_value
        if custom_attrs:
            metadata["custom_attrs"] = custom_attrs

        # Dataset metadata
        if hasattr(self, "metadata") and "dataset_metadata" in self.metadata:
            metadata["dataset_metadata"] = self.metadata["dataset_metadata"]

        return metadata

    def _save_weights(self, path):
        """
        Save model weights and tokenizer to disk.

        This method handles the serialization of model weights and tokenizer.
        It attempts multiple saving strategies to ensure compatibility:
        1. Save tokenizer using dill serialization
        2. Try to save base model using save_pretrained() (HuggingFace style)
        3. Save complete state dict as fallback

        Args:
            path (str): Target directory path where weights and tokenizer will be saved.

        Side Effects:
            - Creates 'tokenizer.bin' in the target directory
            - Creates model weight files (via save_pretrained if available)
            - Creates 'pytorch_model.bin' containing complete state dict

        Note:
            - Tokenizer is serialized using dill to preserve all attributes
            - Base model save_pretrained() may fail for custom models (non-fatal)
            - Complete state dict is always saved as backup
        """
        # Save tokenizer
        with open(f"{path}/tokenizer.bin", "wb") as f:
            dill.dump(self.tokenizer, f)

        # Try to save the underlying base model
        try:
            self.model.save_pretrained(f"{path}", safe_serialization=False)
        except AttributeError:
            # Fallback: if the OmniModel subclass provides its own `save_pretrained`, use it
            if hasattr(self, "save_pretrained"):
                try:
                    self.save_pretrained(path, overwrite=True)
                except Exception:
                    pass

        # Save complete state dict including all components
        with open(f"{path}/pytorch_model.bin", "wb") as f:
            torch.save(self.state_dict(), f)

    def save(self, path, overwrite=False, dtype=torch.float16, **kwargs):
        """
        Save the complete model, tokenizer, and metadata to a directory.

        This method performs a comprehensive save operation that includes:
        - Base model configuration files
        - Custom model and dataset class source code (if applicable)
        - Complete metadata including loss function and custom attributes
        - Model weights and tokenizer

        The save process follows a 6-step workflow:
        1. Save base configuration files
        2. Collect comprehensive metadata
        3. Save custom model class source (if user-defined)
        4. Save custom dataset class source (if available)
        5. Save metadata to JSON file
        6. Save model weights and tokenizer

        Args:
            path (str): Target directory path for saving the model.
            overwrite (bool, optional): Whether to overwrite existing directory.
                                       If False, a timestamp will be appended to path.
                                       Defaults to False.
            dtype (torch.dtype, optional): Data type for saving model weights.
                                          Model is temporarily converted to this dtype
                                          during saving. Defaults to torch.float16.
            **kwargs: Additional keyword arguments (reserved for future use).

        Side Effects:
            - Creates target directory if it doesn't exist
            - Temporarily moves model to CPU and converts dtype
            - Restores original device and dtype after saving
            - Prints confirmation message on completion

        Example:
            >>> # Basic save
            >>> model.save("my_model")

            >>> # Save with overwrite
            >>> model.save("my_model", overwrite=True)

            >>> # Save with specific dtype
            >>> model.save("my_model", dtype=torch.float32)

        Note:
            - Model is set to eval mode before saving
            - Original device and dtype are preserved
            - Custom classes are only saved if defined outside framework packages
            - Failures in non-critical steps (e.g., custom class saving) are logged but don't stop the process
        """
        self.eval()

        # Handle path conflicts
        if os.path.exists(path) and not overwrite:
            fprint(
                f"The path {path} already exists, please set overwrite=True to overwrite it. "
                f"Rename the path to {path}_{time.strftime('%Y%m%d_%H%M%S')} to save it with a timestamp."
            )
            path = f"{path}_{time.strftime('%Y%m%d_%H%M%S')}"
        if not os.path.exists(path):
            os.makedirs(path)

        # Store original device and dtype
        _device = self.model.device
        _dtype = self.model.dtype
        self.model.to(dtype).to("cpu")

        # Save tokenizer config
        self.tokenizer.save_pretrained(path)

        # Step 1: Save base files
        self._save_base_files(path)

        # Step 2: Collect metadata
        metadata = self._collect_metadata()

        # Step 3: Save custom model class
        self._save_custom_model_class(path, metadata)

        # Step 4: Save custom dataset class
        self._save_custom_dataset_class(path, metadata)

        # Step 5: Save metadata to JSON
        with open(f"{path}/metadata.json", "w", encoding="utf8") as f:
            json.dump(metadata, f, indent=2)

        # Step 6: Save weights and tokenizer
        self._save_weights(path)

        # Restore original device and dtype
        self.model.to(_dtype).to(_device)
        fprint(f"The model is saved to {path}.")

    # ==================== Load Helper Methods ====================

    def _load_metadata(self, path):
        """
        Load and validate metadata from saved model directory.

        This method reads the metadata.json file and performs validation
        to ensure the saved model matches the current model class.

        Args:
            path (str): Directory path containing the saved model and metadata.json.

        Returns:
            dict: Loaded metadata dictionary containing model information.

        Raises:
            ValueError: If the saved model class doesn't match the current model class.
            FileNotFoundError: If metadata.json is not found in the directory.
            json.JSONDecodeError: If metadata.json is malformed.

        Example:
            >>> metadata = model._load_metadata("checkpoint")
            >>> print(metadata['model_cls'])  # 'OmniModelForSequenceClassification'

        Note:
            - Validates that saved model class matches current class
            - This check ensures type safety when loading models
        """
        with open(f"{path}/metadata.json", "r", encoding="utf8") as f:
            metadata = json.load(f)

        if metadata["model_cls"] != self.__class__.__name__:
            raise ValueError(
                f"The model class in the loaded model is {metadata['model_cls']}, "
                f"but the current model class is {self.__class__.__name__}."
            )

        return metadata

    def _load_config(self, path, **kwargs):
        """
        Load model configuration and check for differences with current config.

        This method loads the saved configuration and compares it with the
        current model's configuration, warning about any differences found.

        Args:
            path (str): Directory path containing the saved model configuration.
            **kwargs: Additional arguments passed to AutoConfig.from_pretrained().

        Returns:
            AutoConfig: Loaded configuration object.

        Side Effects:
            - Prints warnings for any configuration differences found
            - Warnings include the key name and both values (saved vs current)

        Example:
            >>> config = model._load_config("checkpoint", trust_remote_code=True)
            Warning: The value of the key num_labels in the loaded model is 10,
            but the current value is 5.

        Note:
            - trust_remote_code is set to True by default for custom models
            - Configuration differences don't prevent loading but are logged
            - Useful for detecting model version mismatches
        """
        config = AutoConfig.from_pretrained(path, trust_remote_code=True, **kwargs)

        for key, value in config.__dict__.items():
            if key not in self.config.__dict__ or self.config.__dict__[key] != value:
                fprint(
                    f"Warning: The value of the key {key} in the loaded model is {value}, "
                    f"but the current value is {self.config.__dict__.get(key, None)}."
                )

        return config

    def _load_dataset_class(self, path, metadata):
        """
        Restore the dataset class from metadata.

        This method attempts to dynamically import the dataset class that was
        saved with the model. This allows the model to use the same data
        preprocessing logic during inference as was used during training.

        Args:
            path (str): Directory path containing the saved model.
            metadata (dict): Metadata dictionary containing dataset class information.
                           Expected keys: 'dataset_cls', 'dataset_module', 'custom_dataset_file', 'custom_dataset_class'

        Side Effects:
            - Sets self.dataset_class to the restored dataset class
            - Prints confirmation message on success
            - Prints warning on failure (non-fatal)

        Note:
            - Silently returns if dataset class info is not in metadata
            - Tries multiple loading strategies: built-in modules, custom files
            - Import or attribute errors are caught and logged as warnings
            - Custom dataset classes are loaded from custom_dataset.py if available
        """
        if "dataset_cls" not in metadata and "custom_dataset_class" not in metadata:
            return

        dataset_cls = None

        # Method 1: Try to load from custom_dataset.py file
        if "custom_dataset_file" in metadata and "custom_dataset_class" in metadata:
            try:
                custom_dataset_path = os.path.join(
                    path, metadata["custom_dataset_file"]
                )
                if os.path.exists(custom_dataset_path):
                    import importlib.util

                    spec = importlib.util.spec_from_file_location(
                        "custom_dataset_module", custom_dataset_path
                    )
                    custom_dataset_module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(custom_dataset_module)
                    dataset_cls = getattr(
                        custom_dataset_module, metadata["custom_dataset_class"]
                    )
                    fprint(
                        f"Restored custom dataset class: {metadata['custom_dataset_class']} from {metadata['custom_dataset_file']}"
                    )
            except (ImportError, AttributeError, OSError) as e:
                warnings.warn(f"Could not restore custom dataset class from file: {e}")

        # Method 2: Try to load from built-in omnigenbench/omnigenome modules
        if (
            dataset_cls is None
            and "dataset_module" in metadata
            and "dataset_cls" in metadata
        ):
            try:
                dataset_module = import_module(metadata["dataset_module"])
                dataset_cls = getattr(dataset_module, metadata["dataset_cls"])
                fprint(
                    f"Restored dataset class: {metadata['dataset_cls']} from {metadata['dataset_module']}"
                )
            except (ImportError, AttributeError) as e:
                warnings.warn(f"Could not restore dataset class from module: {e}")

        if dataset_cls is not None:
            self.dataset_class = dataset_cls

    def _load_loss_function(self, metadata):
        """
        Restore saved loss function from metadata.

        This method attempts to restore the loss function that was used
        during training by dynamically importing it based on saved metadata.

        Args:
            metadata (dict): Metadata dictionary containing loss function information.
                           Expected keys: 'loss_fn_class', 'loss_fn_module'

        Side Effects:
            - Sets self.loss_fn to the restored loss function instance
            - Prints confirmation message on success
            - Prints warning on failure (non-fatal)

        Example:
            >>> # metadata contains: {'loss_fn_class': 'CrossEntropyLoss', 'loss_fn_module': 'torch.nn.modules.loss'}
            >>> model._load_loss_function(metadata)
            Restored loss function: CrossEntropyLoss from torch.nn.modules.loss

        Note:
            - Silently returns if loss function info is not in metadata
            - Import or attribute errors are caught and logged as warnings
            - Loss function is instantiated with default parameters
            - Custom loss functions must be importable at load time
        """
        if "loss_fn_class" not in metadata or "loss_fn_module" not in metadata:
            return

        try:
            loss_module = import_module(metadata["loss_fn_module"])
            loss_class = getattr(loss_module, metadata["loss_fn_class"])
            self.loss_fn = loss_class()
            fprint(
                f"Restored loss function: {metadata['loss_fn_class']} from {metadata['loss_fn_module']}"
            )
        except (ImportError, AttributeError) as e:
            warnings.warn(f"Could not restore loss function: {e}")

    def _load_weights(self, path, **kwargs):
        """
        Load model weights with validation and compatibility checks.

        This method loads the saved state dictionary and performs thorough
        validation by comparing saved weights with the current model structure.
        It reports any missing or unexpected keys.

        Args:
            path (str): Directory path containing pytorch_model.bin file.
            **kwargs: Additional arguments, may include:
                     - device (str): Device to map loaded tensors to (e.g., 'cpu', 'cuda:0')

        Side Effects:
            - Loads weights into current model's state dict
            - Prints warnings for missing or unexpected keys
            - Uses strict=False to allow partial loading

        Warnings:
            - Missing keys: Parameters in current model not found in saved weights
            - Unexpected keys: Parameters in saved weights not found in current model

        Example:
            >>> model._load_weights("checkpoint", device="cuda:0")
            Warning: Missing keys in loaded weights: {'classifier.bias'}
            Warning: Unexpected keys in loaded weights: {'old_layer.weight'}

        Note:
            - strict=False allows loading with architecture mismatches
            - Missing keys will be randomly initialized
            - Unexpected keys are ignored
            - Device mapping prevents CUDA OOM when loading on different device
        """
        with open(f"{path}/pytorch_model.bin", "rb") as f:
            loaded_state_dict = torch.load(f, map_location=kwargs.get("device", "cpu"))

            # Check if keys match between current and loaded state dict
            current_keys = set(self.state_dict().keys())
            loaded_keys = set(loaded_state_dict.keys())
            missing_keys = current_keys - loaded_keys
            unexpected_keys = loaded_keys - current_keys

            if missing_keys:
                warnings.warn(f"Missing keys in loaded weights: {missing_keys}")
            if unexpected_keys:
                warnings.warn(f"Unexpected keys in loaded weights: {unexpected_keys}")

            self.load_state_dict(loaded_state_dict, strict=False)

    def _load_tokenizer(self, path):
        """
        Load saved tokenizer from binary file.

        This method deserializes the tokenizer that was saved using dill.
        The tokenizer is essential for proper text preprocessing during inference.

        Args:
            path (str): Directory path containing tokenizer.bin file.

        Side Effects:
            - Sets self.tokenizer to the loaded tokenizer instance
            - Silently returns if tokenizer.bin doesn't exist

        Example:
            >>> model._load_tokenizer("checkpoint")
            >>> print(type(model.tokenizer))  # <class 'omnigenbench.OmniTokenizer'>

        Note:
            - Tokenizer is saved/loaded using dill for complete serialization
            - If tokenizer.bin doesn't exist, current tokenizer is preserved
            - Dill is used instead of pickle to handle complex tokenizer objects
        """
        tokenizer_path = f"{path}/tokenizer.bin"
        if os.path.exists(tokenizer_path):
            with open(tokenizer_path, "rb") as f:
                self.tokenizer = dill.load(f)

    def load(self, path, **kwargs):
        """
        Load a complete model from a saved directory.

        This method performs a comprehensive load operation that restores:
        - Model metadata and configuration
        - Loss function (if saved)
        - Model weights with validation
        - Tokenizer

        The load process follows a 6-step workflow:
        1. Load and validate metadata
        2. Load model configuration
        3. Restore dataset class (if available)
        4. Restore loss function (if available)
        5. Load model weights with validation
        6. Load tokenizer

        Args:
            path (str): Directory path containing the saved model.
            **kwargs: Additional keyword arguments passed to loading functions.
                     Common options include:
                     - device (str): Device to load model to (e.g., 'cpu', 'cuda:0')
                     - trust_remote_code (bool): Whether to trust custom code

        Returns:
            OmniModel: The loaded model instance (self).

        Raises:
            ValueError: If saved model class doesn't match current class.
            FileNotFoundError: If required files (metadata.json, pytorch_model.bin) are missing.

        Example:
            >>> # Basic load
            >>> loaded_model = model.load("checkpoint")

            >>> # Load to specific device
            >>> loaded_model = model.load("checkpoint", device="cuda:0")

            >>> # Load with custom code trust
            >>> loaded_model = model.load("checkpoint", trust_remote_code=True)

        Side Effects:
            - Updates all model attributes (weights, config, tokenizer, loss_fn)
            - Prints warnings for configuration differences
            - Prints warnings for weight loading issues (missing/unexpected keys)
            - Prints confirmation messages for restored components

        Note:
            - Model class must match the saved model class
            - Partial loading is supported (missing keys will be randomly initialized)
            - Loss function restoration is optional (warnings only if fails)
            - Custom models require trust_remote_code=True (enabled by default)
        """
        # Step 1: Load metadata
        metadata = self._load_metadata(path)

        # Step 2: Load configuration
        config = self._load_config(path, **kwargs)

        # Step 3: Restore dataset class
        self._load_dataset_class(path, metadata)

        # Step 4: Restore loss function
        self._load_loss_function(metadata)

        # Step 5: Load weights
        self._load_weights(path, **kwargs)

        # Step 6: Load tokenizer
        self._load_tokenizer(path)

        return self

    def _forward_from_raw_input(self, sequence_or_inputs, **kwargs):
        """
        Tokenizes raw input and performs a forward pass in no_grad mode.

        This method supports two preprocessing strategies:
        1. If dataset_class is available, use dataset's prepare_input method for preprocessing
        2. Otherwise, use direct tokenizer (backward compatible)

        :param sequence_or_inputs: A sequence (str), list of sequences, dict with 'sequence' key,
                                  or tokenized inputs (BatchEncoding/dict).
        :param kwargs: Additional arguments for tokenization or dataset preparation.
        :return: A dictionary containing the raw model outputs and the tokenized inputs.
        """
        # Check if inputs are already tokenized
        if isinstance(sequence_or_inputs, (BatchEncoding, dict)):
            # If it's a dict, check if it contains 'input_ids' (tokenized) or raw data
            if (
                isinstance(sequence_or_inputs, dict)
                and "input_ids" in sequence_or_inputs
            ):
                inputs = sequence_or_inputs
            # If it's a dict without 'input_ids', it might be raw data for dataset.prepare_input
            elif isinstance(sequence_or_inputs, dict) and hasattr(
                self, "dataset_class"
            ):
                # Use dataset's prepare_input method
                try:
                    # Create a temporary dataset instance for using prepare_input
                    max_length = kwargs.pop("max_length", 1024)
                    dataset_instance = self.dataset_class(
                        dataset_name_or_path=None,
                        tokenizer=self.tokenizer,
                        max_length=max_length,
                        **kwargs,
                    )
                    inputs = dataset_instance.prepare_input(
                        sequence_or_inputs, **kwargs
                    )
                    # Remove batch dimension if present (prepare_input may add it)
                    for key, value in inputs.items():
                        if (
                            isinstance(value, torch.Tensor)
                            and value.dim() > 1
                            and value.size(0) == 1
                        ):
                            inputs[key] = value.squeeze(0)
                except Exception as e:
                    warnings.warn(
                        f"Failed to use dataset.prepare_input: {e}. Falling back to tokenizer."
                    )
                    # Fallback to tokenizer for dict input
                    if "sequence" in sequence_or_inputs or "seq" in sequence_or_inputs:
                        seq = sequence_or_inputs.get(
                            "sequence", sequence_or_inputs.get("seq")
                        )
                        inputs = self.tokenizer(
                            seq,
                            padding=kwargs.pop("padding", True),
                            max_length=kwargs.pop("max_length", 1024),
                            truncation=kwargs.pop("truncation", True),
                            return_tensors=kwargs.pop("return_tensors", "pt"),
                            **kwargs,
                        )
                    else:
                        inputs = sequence_or_inputs
            else:
                inputs = sequence_or_inputs
        # Handle string or list of strings
        elif isinstance(sequence_or_inputs, (str, list)):
            # If dataset_class is available, try to use its prepare_input method
            if hasattr(self, "dataset_class"):
                try:
                    # Prepare instance(s) for dataset.prepare_input
                    if isinstance(sequence_or_inputs, str):
                        instance = sequence_or_inputs
                    else:
                        # For list of sequences, we'll process them one by one
                        instance = sequence_or_inputs

                    max_length = kwargs.pop("max_length", 1024)
                    dataset_instance = self.dataset_class(
                        dataset_name_or_path=None,
                        tokenizer=self.tokenizer,
                        max_length=max_length,
                        **kwargs,
                    )

                    if isinstance(instance, list):
                        # Process list of sequences
                        batch_inputs = []
                        for seq in instance:
                            inp = dataset_instance.prepare_input(seq, **kwargs)
                            batch_inputs.append(inp)
                        # Stack all inputs
                        inputs = {
                            key: torch.stack(
                                [
                                    (
                                        inp[key].squeeze(0)
                                        if inp[key].dim() > 0
                                        else inp[key]
                                    )
                                    for inp in batch_inputs
                                ]
                            )
                            for key in batch_inputs[0].keys()
                        }
                    else:
                        inputs = dataset_instance.prepare_input(instance, **kwargs)
                        # Remove batch dimension if present
                        for key, value in inputs.items():
                            if (
                                isinstance(value, torch.Tensor)
                                and value.dim() > 1
                                and value.size(0) == 1
                            ):
                                inputs[key] = value.squeeze(0)
                except Exception as e:
                    warnings.warn(
                        f"Failed to use dataset.prepare_input: {e}. Falling back to tokenizer."
                    )
                    # Fallback to tokenizer
                    inputs = self.tokenizer(
                        sequence_or_inputs,
                        padding=kwargs.pop("padding", True),
                        max_length=kwargs.pop("max_length", 1024),
                        truncation=kwargs.pop("truncation", True),
                        return_tensors=kwargs.pop("return_tensors", "pt"),
                        **kwargs,
                    )
            else:
                # No dataset_class, use tokenizer directly (backward compatible)
                inputs = self.tokenizer(
                    sequence_or_inputs,
                    padding=kwargs.pop("padding", True),
                    max_length=kwargs.pop("max_length", 1024),
                    truncation=kwargs.pop("truncation", True),
                    return_tensors=kwargs.pop("return_tensors", "pt"),
                    **kwargs,
                )
        else:
            raise ValueError(f"Unsupported input type: {type(sequence_or_inputs)}")

        # Ensure inputs are on the correct device and add batch dimension if needed
        if not isinstance(inputs, (BatchEncoding, dict)):
            raise ValueError(f"Processed inputs must be a dict, got {type(inputs)}")

        # Add batch dimension if missing
        for key, value in inputs.items():
            if isinstance(value, torch.Tensor) and value.dim() == 1:
                inputs[key] = value.unsqueeze(0)

        inputs = {
            k: v.to(self.model.device) if isinstance(v, torch.Tensor) else v
            for k, v in inputs.items()
        }

        with torch.no_grad():
            raw_outputs = self(**inputs)
            raw_outputs["inputs"] = inputs
        return raw_outputs

    @staticmethod
    def from_pretrained(model_name_or_path, tokenizer, *args, **kwargs):
        """
        Loads a pre-trained model and tokenizer.

        :param model_name_or_path: The name or path of the pre-trained model.
        :param tokenizer: The tokenizer to use.
        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.
        :return: An instance of `OmniModel`.
        """
        config = kwargs.pop("config", None)
        if config is None:
            config = AutoConfig.from_pretrained(model_name_or_path, **kwargs)
        base_model = AutoModel.from_pretrained(model_name_or_path, **kwargs)
        if tokenizer is None:
            tokenizer = AutoTokenizer.from_pretrained(base_model, **kwargs)
        return OmniModel(config, base_model, tokenizer, *args, **kwargs)

    def model_info(self):
        """
        Prints and returns detailed information about the model.

        :return: A string containing the model information.
        """
        info = f"Model Name: {self.__class__.__name__}\n"
        info += f"Model Metadata: {self.metadata}\n"
        info += f"Base Model Name: {self.config.name_or_path}\n"
        info += f"Model Type: {self.config.model_type}\n"
        info += f"Model Architecture: {self.config.architectures}\n"
        info += f"Model Parameters: {count_parameters(self.model) / 1e6} M\n"
        info += f"Model Config: {self.config}\n"
        fprint(info)
        return info



# -*- coding: utf-8 -*-
# File: dataset_hub.py
# Time: 02:22 20/06/2025
# Author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# Website: https://yangheng95.github.io
# GitHub: https://github.com/yangheng95
# HuggingFace: https://huggingface.co/yangheng
# Google Scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2025. All rights reserved.
"""
This module provides utilities for loading benchmark datasets from the OmniGenome hub.
It handles automatic downloading, configuration loading, and dataset initialization
for various genomic benchmarks.
"""

import os
import warnings

import findfile
from typing_extensions import Union

from ...abc.abstract_tokenizer import OmniTokenizer
from ...misc.utils import load_module_from_path, fprint
from ..hub_utils import download_benchmark


def load_benchmark_datasets(
    benchmark: str,
    tokenizer: Union["OmniTokenizer", str] = None,
    **kwargs: dict,
):
    """
    This function automatically downloads benchmark datasets if they don't exist locally,
    loads their configurations, and initializes train/validation/test datasets with
    the specified tokenizer.

    Args:
        benchmark (str): Name or path of the benchmark to load. If the benchmark
            doesn't exist locally, it will be downloaded from the hub.
        tokenizer (Union[OmniTokenizer, str], optional): Tokenizer to use for
            dataset preprocessing. Can be an OmniTokenizer instance or a string
            identifier for a pre-trained tokenizer. If None, the tokenizer will
            be loaded from the benchmark configuration.
        **kwargs: Additional keyword arguments to override benchmark configuration.
            These will be passed to the dataset classes and tokenizer initialization.

    Returns:
        dict: Dictionary containing datasets for each benchmark task, with keys being benchmark names and values being dictionaries with 'train', 'valid', and 'test' datasets.

    Raises:
        FileNotFoundError: If the benchmark cannot be found or downloaded.
        ValueError: If the benchmark configuration is invalid.
        ImportError: If required dependencies are not available.

    Example:
        >>> from omnigenbench import OmniSingleNucleotideTokenizer
        >>> tokenizer = OmniSingleNucleotideTokenizer.from_pretrained("model_name")
        >>> datasets = load_benchmark_datasets("RGB", tokenizer, max_length=512)
        >>> print(f"Loaded {len(datasets)} benchmark tasks")
        >>> for task_name, task_datasets in datasets.items():
        ...     print(f"{task_name}: {len(task_datasets['train'])} train samples")

    Note:
        - The function automatically handles U/T conversion and other preprocessing
          based on the benchmark configuration.
        - If a tokenizer string is provided, it will be loaded with the benchmark's
          trust_remote_code setting.
        - The function supports multiple seeds for robust evaluation.
        - Long sequences can be dropped or truncated based on configuration.
    """

    benchmark = download_benchmark(benchmark)

    # Import benchmark list
    bench_metadata = load_module_from_path(
        f"bench_metadata", f"{benchmark}/metadata.py"
    )
    datasets = {}
    for _, bench in enumerate(bench_metadata.bench_list):
        bench_config_path = findfile.find_file(
            benchmark, f"{benchmark}.{bench}.config".split(".")
        )
        config = load_module_from_path("config", bench_config_path)
        bench_config = config.bench_config
        fprint(f"Loaded config for {bench} from {bench_config_path}")
        fprint(bench_config)
        _kwargs = kwargs.copy()

        # Init Tokenizer and Model
        if isinstance(tokenizer, str):
            tokenizer = OmniTokenizer.from_pretrained(
                tokenizer,
                trust_remote_code=bench_config.get("trust_remote_code", True),
                **bench_config,
            )

        for key, value in _kwargs.items():
            if key in bench_config:
                fprint("Override", key, "with", value, "according to the input kwargs")
                bench_config.update({key: value})

            else:
                warnings.warn(
                    f"kwarg: {key} not found in bench_config while setting {key} = {value}"
                )
                bench_config.update({key: value})

        for key, value in bench_config.items():
            if key in bench_config and key in _kwargs:
                _kwargs.pop(key)

        if not isinstance(bench_config["seeds"], list):
            bench_config["seeds"] = [bench_config["seeds"]]

        # Init Trainer
        dataset_cls = bench_config["dataset_cls"]

        max_length = bench_config["max_length"]

        train_set = dataset_cls(
            dataset_name_or_path=bench_config["train_file"],
            tokenizer=tokenizer,
            label2id=bench_config["label2id"],
            max_length=max_length,
            structure_in=bench_config.get("structure_in", False),
            max_examples=bench_config.get("max_examples", None),
            shuffle=bench_config.get("shuffle", True),
            drop_long_seq=bench_config.get("drop_long_seq", False),
            **_kwargs,
        )
        test_set = dataset_cls(
            dataset_name_or_path=bench_config["test_file"],
            tokenizer=tokenizer,
            label2id=bench_config["label2id"],
            max_length=max_length,
            structure_in=bench_config.get("structure_in", False),
            max_examples=bench_config.get("max_examples", None),
            shuffle=False,
            drop_long_seq=bench_config.get("drop_long_seq", False),
            **_kwargs,
        )
        valid_set = dataset_cls(
            dataset_name_or_path=bench_config["valid_file"],
            tokenizer=tokenizer,
            label2id=bench_config["label2id"],
            max_length=max_length,
            structure_in=bench_config.get("structure_in", False),
            max_examples=bench_config.get("max_examples", None),
            shuffle=False,
            drop_long_seq=bench_config.get("drop_long_seq", False),
            **_kwargs,
        )

        dataset = {
            "train": train_set,
            "test": test_set,
            "valid": valid_set,
        }

        fprint(
            f"Loaded dataset for {bench} with {len(train_set)} train samples, "
            f"{len(test_set)} test samples and {len(valid_set)} valid samples."
        )

        datasets[bench] = dataset

    return datasets



# -*- coding: utf-8 -*-
# file: auto_bench_config.py
# time: 14:58 29/04/2024
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2024. All Rights Reserved.
from argparse import Namespace

from ...auto.auto_bench.config_check import config_check


class AutoConfig:
    """
    A configuration class for AutoBench and AutoTrain from OmniGenBench.

    This class holds the configuration parameters for a benchmark run. It behaves
    like a dictionary and also tracks how many times each parameter is accessed.
    """

    def __init__(self, args=None, **kwargs):
        """
        Initializes the AutoConfig.

        :param args: A dictionary or `argparse.Namespace` of parameters.
        :param kwargs:
        """
        if not args:
            args = {}
        super().__init__(**kwargs)

        if isinstance(args, Namespace):
            self.args = vars(args)
            self.args_call_count = {arg: 0 for arg in vars(args)}
        else:
            self.args = args
            self.args_call_count = {arg: 0 for arg in args}

    def __getattribute__(self, arg_name):
        """
        Get the value of an argument and increment its call count.

        :param arg_name: The name of the argument.
        :return: The value of the argument.
        """
        if arg_name == "args" or arg_name == "args_call_count":
            return super().__getattribute__(arg_name)
        try:
            value = super().__getattribute__("args")[arg_name]
            args_call_count = super().__getattribute__("args_call_count")
            args_call_count[arg_name] += 1
            super().__setattr__("args_call_count", args_call_count)
            return value

        except Exception as e:
            return super().__getattribute__(arg_name)

    def __setattr__(self, arg_name, value):
        """
        Set the value of an argument and add it to the argument dict and call count dict.

        :param arg_name: The name of the argument.
        :param value: The value of the argument.
        """
        if arg_name == "args" or arg_name == "args_call_count":
            super().__setattr__(arg_name, value)
            return
        try:
            args = super().__getattribute__("args")
            args[arg_name] = value
            super().__setattr__("args", args)
            args_call_count = super().__getattribute__("args_call_count")

            if arg_name in args_call_count:
                super().__setattr__("args_call_count", args_call_count)

            else:
                args_call_count[arg_name] = 0
                super().__setattr__("args_call_count", args_call_count)

        except Exception as e:
            super().__setattr__(arg_name, value)

    def get(self, key, default=None):
        """
        Get the value of a key from the parameter dict. If the key is found, increment its call frequency.
        :param key: The key to look for in the parameter dict.
        :param default: The default value to return if the key is not found.
        :return: The value of the key in the parameter dict, or the default value if the key is not found.
        """
        if key in self.args_call_count:
            self.args_call_count[key] += 1
        return self.args.get(key, default)

    def update(self, *args, **kwargs):
        """
        Update the parameter dict with the given arguments and keyword arguments, and check if the updated configuration is valid.
        :param args: Positional arguments to update the parameter dict.
        :param kwargs: Keyword arguments to update the parameter dict.
        """
        self.args.update(*args, **kwargs)
        config_check(self.args)

    def pop(self, *args):
        """
        Pop a value from the parameter dict.
        :param args: Arguments to pop from the parameter dict.
        :return: The value popped from the parameter dict.
        """
        return self.args.pop(*args)

    def keys(self):
        """
        Get a list of all keys in the parameter dict.
        :return: A list of all keys in the parameter dict.
        """
        return self.args.keys()

    def values(self):
        """
        Get a list of all values in the parameter dict.
        :return: A list of all values in the parameter dict.
        """
        return self.args.values()

    def items(self):
        """
        Get a list of all key-value pairs in the parameter dict.
        :return: A list of all key-value pairs in the parameter dict.
        """
        return self.args.items()

    def __str__(self):
        """
        Get a string representation of the parameter dict.
        :return: A string representation of the parameter dict.
        """
        return str(self.args)

    def __repr__(self):
        """
        Return a detailed string representation of the configuration,
        including all parameters and the frequency of their access.
        """
        param_list = []
        for key, value in self.args.items():
            count = self.args_call_count.get(key, 0)
            param_list.append(f"{key}={value!r} (accessed {count} times)")
        params_str = ", ".join(param_list)
        return f"{self.__class__.__name__}({params_str})"

    def __len__(self):
        """
        Return the number of items in the parameter dict.
        """
        return len(self.args)

    def __iter__(self):
        """
        Return an iterator over the keys of the parameter dict.
        """
        return iter(self.args)

    def __contains__(self, item):
        """
        Check if the given item is in the parameter dict.
        :param item: The item to check.
        :return: True if the item is in the parameter dict, False otherwise.
        """
        return item in self.args

    def __getitem__(self, item):
        """
        Get the value of a key from the parameter dict.
        :param item: The key to look for in the parameter dict.
        :return: The value of the key in the parameter dict.
        """
        return self.args[item]

    def __setitem__(self, key, value):
        """
        Set the value of a key in the parameter dict. Also set the call frequency of the key to 0 and check if the updated
        configuration is valid.
        :param key: The key to set the value for in the parameter dict.
        :param value: The value to set for the key in the parameter dict.
        """
        self.args[key] = value
        self.args_call_count[key] = 0
        config_check(self.args)

    def __delitem__(self, key):
        """
        Delete a key-value pair from the parameter dict and check if the updated configuration is valid.
        :param key: The key to delete from the parameter dict.
        """
        del self.args[key]
        config_check(self.args)

    def __eq__(self, other):
        """
        Check if the parameter dict is equal to another object.
        :param other: The other object to compare with the parameter dict.
        :return: True if the parameter dict is equal to the other object, False otherwise.
        """
        return self.args == other

    def __ne__(self, other):
        """
        Check if the parameter dict is not equal to another object.
        :param other: The other object to compare with the parameter dict.
        :return: True if the parameter dict is not equal to the other object, False otherwise.
        """
        return self.args != other


class AutoBenchConfig(AutoConfig):
    pass
